/* fitgauss -- fit Gaussian to section of image */

/* To compile:
       gcc -o fitgauss fitgauss.c -L. -lcfitsio -lm -lnsl -lsocket */
/* Also requires:  fitsio.h, libcfitsio.a, longnam.h */

/* C. Darren Dowell */

/* version 1.0 -- 2003 Feb 8 */
/* version 1.1 -- 2003 Jun 25 -- updated for new file format
                                 stderr and stdout output */
/* version 2.0 -- 2005 Oct 30 -- one more decimal place; verbose mode;
                                 don't require weight image;
                                 better handling of undefined pixels */
/* version 2.1 -- 2005 Dec  3 -- pointing support for CRUSH images
                                 use adjusted pointing data, if available */
/* version 2.2 -- 2006 May 28 -- fit over circular aperture
                                 pixels to fit slide during minimization */
/* version 2.3 -- 2006 Jul  4 -- support SHARP quicklook images */
/* version 3.0 -- 2006 Jul 27 -- add 'e' option (elliptical gaussian)
                                 add 'ix0', 'iy0', 'fsig'
                                 change meaning of 'sig' */
/* version 3.1 -- 2007 Feb 10 -- update for sharpinteg */
/* version 3.2 -- 2007 Mar 14 -- add 'b' option (background subtraction)
                                 add 'p' option (start on peak) */
/* version 3.3 -- 2007 Sep 12 -- don't require certain image type
                                 check {CD1_2,}C2_2 if CDELT2 not available */

/* SHOULD CONSIDER USING NOISE IMAGE, AS DOES PSFILTER */

/* CRUSH images don't contain FAZO, FZAO */

/* Usage:  fitgauss image.fits {options} */

#define  _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <math.h>
#include "fitsio.h"
#include "nrutil.h"

#define  MAX_LEN  1000
#define  X0  101.0
#define  Y0  101.0
#define  SCALE  2.425
#define  SIG0  1.6
#define  PI  3.141592653589
#define  NUMSIG  5.0

float chi2(float fitparam[]);
float chi2_elliptical(float fitparam[]);
float alpha(float fitparam[]);
float alpha_elliptical(float fitparam[]);
float meanweight(float fitparam[]);
void printabort(char *message);
void printerror(int status);

/* global variables passed to chi2 function */
long  naxes[2];
float  *image, *weight;
char  *nullarray;
int  xmin, xmax, ymin, ymax;
float  sum0;
float  delta;

int main(int argc, char *argv[])

{
    char  version[] = "fitgauss 3.3, 2007 Sep 11";
    char  line[MAX_LEN], comment[MAX_LEN];
    fitsfile  *infptr;
    int  status = 0;     /* must be initialized to zero! */
    int  naxis, bitpix;
    long  pixel0[2], nelements;
    int  anynull;
    float  nullvalue = FLOATNULLVALUE;
    int  hdutype;
    float  p[8], **xi;
    float  ftol, fret;
    int  iter;
    int  i, j;
    int  n;
    void powell();
    float  sigma2, sigmaa;
    int  arg;
    float  x0, y0, scale, scale1, scale2;
    int  x, y;
    float  peak;
    float  fazo, fzao, temp, dx, dy, fazo_new, fzao_new;
    float  azo, zao;
    char  altaz[MAX_LEN];
    char  verbose = 'n', sharp = 'n', bgnd = 'n', startpeak = 'n';
    float  term;
    float  sharp_rot = 0.0, ssr, csr;
    char  elliptical = 'n';
    float  phi, sin2phi, a, b;
    float  sumb, sum1;
    int  xpeak, ypeak;


    fprintf(stderr, "%s\n", version);
    if (argc < 2)
      printabort("Usage:  fitgauss image.fits {{f}sigma #, {f,i}x0 #, {f,i}y0 #, v, e, b, p}");
    if (fits_open_file(&infptr, argv[1], READONLY, &status))
      printerror(status);
    if (fits_get_img_param(infptr, 2, &bitpix, &naxis, naxes, &status))
          printerror(status);
    if (naxis != 2)
      printabort("Image must be 2 dimensional");
    if (fits_read_key(infptr, TFLOAT, "CRPIX1", &x0, comment, &status)) {
      status = 0;
      x0 = X0;
    }
    if (fits_read_key(infptr, TFLOAT, "CRPIX2", &y0, comment, &status)) {
      status = 0;
      y0 = Y0;
    }
    if (fits_read_key(infptr, TFLOAT, "CDELT2", &scale, comment, &status)) {
      status = 0;
      if (fits_read_key(infptr, TFLOAT, "CD2_2", &scale2, comment, &status)) {
        status = 0;
        scale = SCALE;
      }
      else {
        if (fits_read_key(infptr, TFLOAT, "CD1_2", &scale1, comment, &status)) {
          status = 0;
          scale1 = 0.0;
        }
        scale = sqrt(scale1*scale1 + scale2*scale2)*3600.0;
      }
    }
    else {
      scale *= 3600.0;
    }

    pixel0[0] = 1;
    pixel0[1] = 1;
    nelements = naxes[0]*naxes[1];
    image = (float *) malloc(nelements*sizeof(float));
    nullarray = (char *) malloc(nelements*sizeof(char));
/*  if (fits_read_pix(infptr, TFLOAT, pixel0, nelements, &nullvalue,
                      image, &anynull, &status)) */
    if (fits_read_pixnull(infptr, TFLOAT, pixel0, nelements, image,
                      nullarray, &anynull, &status))
      printerror(status);
    weight = (float *) malloc(nelements*sizeof(float));
    if (fits_movabs_hdu(infptr, 2, &hdutype, &status)) {
      status = 0;
      for (i = 0; i < nelements; i++)
        weight[i] = 1.0;
    }
    else {
      if (fits_read_pix(infptr, TFLOAT, pixel0, nelements, &nullvalue,
                        weight, &anynull, &status))
        printerror(status);
    }
    for (i = 0; i < nelements; i++)
      if (nullarray[i] == 1)
        weight[i] = 0.0;
/*  if (fits_movabs_hdu(infptr, 5, &hdutype, &status)) { */
    if (fits_movnam_hdu(infptr, 0, "Primary", 0, &status)) {
      status = 0;
    }
    if (fits_read_key(infptr, TFLOAT, "FAZO", &fazo, comment, &status)) {
      status = 0;
      fazo = sqrt(-1.0);     /* force NaN */
    }
    if (fits_read_key(infptr, TFLOAT, "FZAO", &fzao, comment, &status)) {
      status = 0;
      fzao = sqrt(-1.0);
    }
    if (fits_read_key(infptr, TFLOAT, "AZO", &azo, comment, &status)) {
      status = 0;
      azo = 0.0;
    }
    if (fits_read_key(infptr, TFLOAT, "ZAO", &zao, comment, &status)) {
      status = 0;
      azo = 0.0;
    }
    if (fits_movabs_hdu(infptr, 1, &hdutype, &status))
      printerror(status);
    if (fits_read_key(infptr, TFLOAT, "FAZON001", &temp, comment, &status)) {
      status = 0;
    }
    else {
      fazo = temp;
    }
    if (fits_read_key(infptr, TFLOAT, "FZAON001", &temp, comment, &status)) {
      status = 0;
    }
    else {
      fzao = temp;
    }
    if (fits_read_key(infptr, TFLOAT, "TEL_ELEV", &temp, comment, &status)) {
      status = 0;
    }
    else {
      sharp_rot = temp;
      if (fits_read_key(infptr, TFLOAT, "DEWARANG", &temp,comment,&status)) {
        status = 0;
      }
      else {
        sharp = 'y';
        sharp_rot -= temp;     /* elev - ROTATOR_0, as in sharcsolve */
      }
    }
    if (fits_read_key(infptr, TSTRING, "ALTAZ", altaz, comment, &status)) {
      status = 0;
      if (sharp != 'y') {
        fazo = sqrt(-1.0);
        fzao = sqrt(-1.0);
      }
    }
    else {
      sharp = 'n';
      if ((altaz[0] != 'y') && (altaz[0] != 'Y') && (altaz[0] != 'T')) {
        fazo = sqrt(-1.0);
        fzao = sqrt(-1.0);
      }
    }

    if (fits_close_file(infptr, &status))
      printerror(status);

    xi = matrix(1, 7, 1, 7);
    for (i = 1; i <= 7; i++)
      for (j = 1; j <= 7; j++)
        xi[i][j] = 0.0;
    xi[1][1] = 1.0;
    xi[2][2] = 1.0;
    xi[3][3] = 1.0;
    xi[4][4] = 1.0;
    xi[5][5] = 1.0;
    xi[6][6] = 1.0;
    xi[7][7] = 1.0;
    p[1] = 0.0;     /* amplitude */
    p[2] = SIG0;    /* sigma(x) */
    p[3] = 0.0;     /* background */
    p[4] = x0;      /* x center */
    p[5] = y0;      /* y center */
    p[6] = SIG0;    /* sigma y */
    p[7] = 0.0;     /* elliptical gaussian rotation term */

    ftol = 1.0e-6;
    arg = 2;
    while (arg < argc) {
      if (strncmp(argv[arg], "si", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &(p[2])) != 1)
          printabort("Didn't understand sigma");
        p[6] = p[2];
      }
      if (strncmp(argv[arg], "fsi", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &(p[2])) != 1)
          printabort("Didn't understand fsigma");
        p[6] = p[2];
        xi[2][2] = 0.0;
        xi[6][6] = 0.0;
        xi[7][7] = 0.0;
      }
      if (strncmp(argv[arg], "x0", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &x0) != 1)
          printabort("Didn't understand x0");
        p[4] = x0;
      }
      if (strncmp(argv[arg], "ix0", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &(p[4])) != 1)
          printabort("Didn't understand ix0");
      }
      if (strncmp(argv[arg], "fx0", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &x0) != 1)
          printabort("Didn't understand fx0");
        p[4] = x0;
        xi[4][4] = 0.0;
      }
      if (strncmp(argv[arg], "y0", 2) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &y0) != 1)
          printabort("Didn't understand y0");
        p[5] = y0;
      }
      if (strncmp(argv[arg], "iy0", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &(p[5])) != 1)
          printabort("Didn't understand iy0");
      }
      if (strncmp(argv[arg], "fy0", 3) == 0) {
        arg++;
        if (sscanf(argv[arg], "%f", &y0) != 1)
          printabort("Didn't understand fy0");
        p[5] = y0;
        xi[5][5] = 0.0;
      }
      if (strncmp(argv[arg], "v", 1) == 0) {
        verbose = 'y';
      }
      if (strncmp(argv[arg], "e", 1) == 0) {
        elliptical = 'y';
      }
      if (strncmp(argv[arg], "b", 1) == 0) {
        bgnd = 'y';
      }
      if (strncmp(argv[arg], "p", 1) == 0) {
        startpeak = 'y';
      }
      arg++;
    }

    peak = -1.0e99;
    sumb = sum1 = 0.0;
    for (x = 1; x <= naxes[0]; x++) {
      for (y = 1; y <= naxes[1]; y++) {
        if (weight[(y-1)*naxes[0] + (x-1)] > 0.1) {
          sum1 += 1.0;
          sumb += image[(y-1)*naxes[0] + (x-1)];
          if (image[(y-1)*naxes[0] + (x-1)] > peak) {
            peak = image[(y-1)*naxes[0] + (x-1)];
            xpeak = x;
            ypeak = y;
          }
        }
      }
    }
    sumb /= sum1;
    if (bgnd == 'y') {
      for (x = 1; x <= naxes[0]; x++) {
        for (y = 1; y <= naxes[1]; y++) {
          if (weight[(y-1)*naxes[0] + (x-1)] > 0.1) {
            image[(y-1)*naxes[0] + (x-1)] -= sumb;
          }
        }
      }
    }
    if (startpeak == 'y') {
      p[4] = xpeak;
      p[5] = ypeak;
    }

    delta = p[2]*NUMSIG;    /* fit radius, fixed during integration */
    if (elliptical == 'y') {
      n = 7;
      powell(p, xi, n, ftol, &iter, &fret, chi2_elliptical);
    }
    else {
      n = 5;
      powell(p, xi, n, ftol, &iter, &fret, chi2);
    }

    peak = -1.0e99;
    /* xmin, ... have been set within chi2 */
    for (x = xmin; x <= xmax; x++) {
      for (y = ymin; y <= ymax; y++) {
        if ((x >= 1) && (x <= naxes[0]) && (y >= 1) && (y <= naxes[1])) {
          term = (x - p[4])*(x - p[4]);
          term += (y - p[5])*(y - p[5]);
          if (term <= (delta*delta)) {
            if (image[(y-1)*naxes[0] + (x-1)] > peak) {
              peak = image[(y-1)*naxes[0] + (x-1)];
              xpeak = x;
              ypeak = y;
            }
          }
        }
      }
    }

    dx = (p[4]-x0)*scale;
    dy = (y0-p[5])*scale;
    if (sharp == 'y') {
      ssr = sin(sharp_rot*PI/180.0);
      csr = cos(sharp_rot*PI/180.0);
      fazo_new = fazo - dx*csr + dy*ssr + azo;
      fzao_new = fzao + dx*ssr + dy*csr + zao;
    }
    else {
      fazo_new = fazo+dx;
      fzao_new = fzao+dy;
    }

    fprintf(stderr, "%d iterations, %.0f pixels, chi2 = %f\n",
            iter, sum0, fret);
    sigma2 = fret/sum0;
    fprintf(stderr, " rms = %.4f, mean weight = %.0f\n", sqrt(sigma2),
            meanweight(p));
    fprintf(stderr, " peak signal = %.4f\n", peak);
    if (elliptical == 'y') {
      sigmaa = sqrt(sigma2/alpha_elliptical(p));
    }
    else {
      sigmaa = sqrt(sigma2/alpha(p));
    }
    fprintf(stderr, " amplitude = %.4f +- %.4f, %.2f sigma\n", p[1], sigmaa,
            p[1]/sigmaa);
    if (elliptical == 'y') {
      phi = atan2(-1.0*p[7], 1.0/p[2]/p[2] - 1.0/p[6]/p[6])/2.0;
      sin2phi = sin(2*phi);
      b = p[7]/sin2phi + 1.0/p[2]/p[2] + 1.0/p[6]/p[6];
      b = sqrt(2.0/b);
      a = 1.0/p[2]/p[2] + 1.0/p[6]/p[6] - 1.0/b/b;
      a = sqrt(1.0/a);
      phi *= (180.0/PI);
      phi *= -1.0;     /* report CCW from +x axis */
      if (b > a) {
        phi += 90.0;
        temp = a;
        a = b;
        b = temp;
      }
      while (phi > 90.0)
        phi -= 180.0;
      while (phi < -90.0)
        phi += 180.0;
      fprintf(stderr, " Gaussian sigma = %.3f, %.3f pix, %.1f deg., FWHM = %.2f, %.2f arcsec\n",
              a, b, phi, a*2.0*sqrt(2.0*log(2.0))*scale,
              b*2.0*sqrt(2.0*log(2.0))*scale);
    }
    else {
      fprintf(stderr, " Gaussian sigma = %.3f pix, FWHM = %.2f arcsec\n", p[2],
              p[2]*2.0*sqrt(2.0*log(2.0))*scale);
    }
    if (elliptical == 'y') {
      fprintf(stderr, " integrated signal = %.4f\n", 2.0*PI*p[1]*p[2]*p[6]);
    }
    else {
      fprintf(stderr, " integrated signal = %.4f\n", 2.0*PI*p[1]*p[2]*p[2]);
    }
    fprintf(stderr, " baseline = %.4f +- %.4f\n", p[3], sqrt(sigma2/sum0));
    fprintf(stderr, " center = %.3f, %.3f\n", p[4], p[5]);
    fprintf(stderr,
            "   position error (right, down), in arcsec:  (%.1f, %.1f)\n",
            dx, dy);
    fprintf(stderr,
            "   suggested (FAZO, FZAO):  (%.1f, %.1f)\n", fazo_new, fzao_new);
    /* One-line summary: */
/*  fprintf(stdout, "%8.3f %5.3f %5.2f %6.1f %6.1f %.1f %.1f\n", p[1], sigmaa,
            p[2]*2.0*sqrt(2.0*log(2.0))*scale, fazo_new, fzao_new, dx, dy); */
    if (elliptical == 'y') {
      fprintf(stdout, "%8.3f %5.3f %5.2f %5.2f %5.1f %6.1f %6.1f", p[1],sigmaa,
              a*2.0*sqrt(2.0*log(2.0))*scale, b*2.0*sqrt(2.0*log(2.0))*scale,
              phi, fazo_new, fzao_new);
    }
    else {
      fprintf(stdout, "%8.3f %5.3f %5.2f %6.1f %6.1f", p[1], sigmaa,
              p[2]*2.0*sqrt(2.0*log(2.0))*scale, fazo_new, fzao_new);
    }
    if (verbose == 'y')
      fprintf(stdout, "   %5.1f %5.1f  %.0f", dx, dy, sum0);
    fprintf(stdout, "\n");
}

/*******************/

float chi2(float fitparam[])

{
    float  term, sum;
    int  x, y;

    sum = 0.0;
    sum0 = 0.0;
    xmin = fitparam[4] - delta - 1.0;     /* sliding fit aperture */
    xmax = fitparam[4] + delta + 1.0;
    ymin = fitparam[5] - delta - 1.0;
    ymax = fitparam[5] + delta + 1.0;
    for (x = xmin; x <= xmax; x++) {
      for (y = ymin; y <= ymax; y++) {
        if ((x >= 1) && (x <= naxes[0]) && (y >= 1) && (y <= naxes[1])) {
          term = (x - fitparam[4])*(x - fitparam[4]);
          term += (y - fitparam[5])*(y - fitparam[5]);
          if (term <= (delta*delta)) {
            term = exp(term/-2.0/fitparam[2]/fitparam[2]);
            term *= fitparam[1];
            term += fitparam[3];
            term -= image[(y-1)*naxes[0] + (x-1)];
            if (weight[(y-1)*naxes[0] + (x-1)] > 0.1) {
              sum += term*term;
              sum0 += 1.0;
            }
          }
        }
      }
    }

    return (sum);
}

/*******************/

float chi2_elliptical(float fitparam[])

{
    float  term, sum;
    int  x, y;

    sum = 0.0;
    sum0 = 0.0;
    xmin = fitparam[4] - delta - 1.0;     /* sliding fit aperture */
    xmax = fitparam[4] + delta + 1.0;
    ymin = fitparam[5] - delta - 1.0;
    ymax = fitparam[5] + delta + 1.0;
    for (x = xmin; x <= xmax; x++) {
      for (y = ymin; y <= ymax; y++) {
        if ((x >= 1) && (x <= naxes[0]) && (y >= 1) && (y <= naxes[1])) {
          term = (x - fitparam[4])*(x - fitparam[4]);
          term += (y - fitparam[5])*(y - fitparam[5]);
          if (term <= (delta*delta)) {
            term = (x - fitparam[4])*(x - fitparam[4])/fitparam[2]/fitparam[2];
            term += (y - fitparam[5])*(y - fitparam[5])/fitparam[6]/fitparam[6];
            term += (x - fitparam[4])*(y - fitparam[5])*fitparam[7];
            term = exp(term/-2.0);
            term *= fitparam[1];
            term += fitparam[3];
            term -= image[(y-1)*naxes[0] + (x-1)];
            if (weight[(y-1)*naxes[0] + (x-1)] > 0.1) {
              sum += term*term;
              sum0 += 1.0;
            }
          }
        }
      }
    }

    return (sum);
}

/*******************/

float alpha_elliptical(float fitparam[])

{
    float  term, sum;
    int  x, y;

    sum = 0.0;
    sum0 = 0.0;
    for (x = xmin; x <= xmax; x++) {
      for (y = ymin; y <= ymax; y++) {
        if ((x >= 1) && (x <= naxes[0]) && (y >= 1) && (y <= naxes[1])) {
          term = (x - fitparam[4])*(x - fitparam[4]);
          term += (y - fitparam[5])*(y - fitparam[5]);
          if (term <= (delta*delta)) {
            term = (x - fitparam[4])*(x - fitparam[4])/fitparam[2]/fitparam[2];
            term += (y - fitparam[5])*(y - fitparam[5])/fitparam[6]/fitparam[6];
            term += (x - fitparam[4])*(y - fitparam[5])*fitparam[7];
            term = exp(term/-2.0);
            if (weight[(y-1)*naxes[0] + (x-1)] > 0.1) {
              sum += term*term;
              sum0 += 1.0;
            }
          }
        }
      }
    }

    return (sum);
}

/*******************/

float alpha(float fitparam[])

{
    float  term, sum;
    int  x, y;

    sum = 0.0;
    sum0 = 0.0;
    for (x = xmin; x <= xmax; x++) {
      for (y = ymin; y <= ymax; y++) {
        if ((x >= 1) && (x <= naxes[0]) && (y >= 1) && (y <= naxes[1])) {
          term = (x - fitparam[4])*(x - fitparam[4]);
          term += (y - fitparam[5])*(y - fitparam[5]);
          if (term <= (delta*delta)) {
            term = exp(term/-2.0/fitparam[2]/fitparam[2]);
            if (weight[(y-1)*naxes[0] + (x-1)] > 0.1) {
              sum += term*term;
              sum0 += 1.0;
            }
          }
        }
      }
    }

    return (sum);
}

/*******************/

float meanweight(float fitparam[])

{
    float  term, sum;
    int  x, y;

    sum = 0.0;
    sum0 = 0.0;
    for (x = xmin; x <= xmax; x++) {
      for (y = ymin; y <= ymax; y++) {
        if ((x >= 1) && (x <= naxes[0]) && (y >= 1) && (y <= naxes[1])) {
          term = (x - fitparam[4])*(x - fitparam[4]);
          term += (y - fitparam[5])*(y - fitparam[5]);
          if (term <= (delta*delta)) {
            if (weight[(y-1)*naxes[0] + (x-1)] > 0.1) {
              sum += weight[(y-1)*naxes[0] + (x-1)];
              sum0 += 1.0;
            }
          }
        }
      }
    }
    if (sum0 > 0.1)
      sum /= sum0;

    return (sum);
}

/*******************/

void printabort(char *message)

{
    fprintf(stderr, "ERROR:  %s\n", message);
    exit (1);
}

/*******************/

void printerror(int status)

{
    if (status)
    {
       fits_report_error(stderr, status);
       exit( status );
    }
    return;
}
