/*Usage: sharpinteg 0hwp.fits 22.5hwp.fits 45hwp.fits 67.5hwp.fits -r 
rgmfile {-c -f # -t # -a # # -q  -o -d str -y # file -nh -nc -sil -fa -hard -ne}*/   
/*-c : Use Chauvenet's theorem for deglitching*/   
/*-f # : Manually set gain between h and v arrays*/    
/*-t  # : Manually set chop tolerance*/   
/*-a # # :  X and Y offsets to align V array with H*/    
/*-q : Quicklook mode outputing h and v array images for each hwp position*/
/*-fa : Print out full array
/*-y :  str file: Progam will read in string str (yyyymmdd) to be used for calculating smoothed tau using tau data from file    
/*-nh :  Do not check hwp angles  
/*-nc  :   Do not check nod/chop agreement   
/*-sil :  Quiet mode (suppresses some messages to screen\n   
/*-o : If this flag is present sharpinteg will output calibration/debug values\n    
/*-d str : Program expects to read in string(date) to be printed to ouput text file generaged by -o flag
/*-hard : Hardware test so do not process chop/nod
/*-ne :  No EDAS
/*-w:  Do Set flux of widow pixels to 2*their flux) */
/*-shn #:  Minimum number of samples required for a hwp integration";*/
/*-em :  Inflate Signal Errors less than median Signal Error to the median*/
/*-et :  Inflate Signal Errors less than the Mean Signal Error - 3*Std Dev Signal Error to the Mean Signal Error*/
/*-m # :  Mask out pixels with Q or U errors above #, in millionths*/ 

/* Version 2 - For use with data files containing all 4 hwp positions 
last edited Nov 8 2006*/
/* Version 2.1 - Added flags for use in hardware tests  last edited Jan 25 2007*/
/* Version 2.2 - Added flag to toggle widow pixels  last edited Mar 08 2007*/
/* Version 2.3 - Added flag to set min # of samples for a hwp last edited Apr 30 2007*/
/* Version 2.4 - Added flags to inflate signal errors last edited Feb 07 2008*/
/* Version 2.4 - Added flags to inflate signal errors last edited Feb 07 2008*/
/* Version 2.5 - Add code to reject data with Q or U errors above .0002 */


#define _GNU_SOURCE
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "nrutil.h"
#include <fitsio.h>
#include <longnam.h>

#define  MASK_VALUE0  10000.0
#define  SCALE_NAS  4.75
#define SCALE_CASS 4.75
#define MAX_HWP 5
#define BLANK NAN
#define  ARRAY_ROWS  12
#define  ARRAY_COLUMNS  32
#define  SHARP_COLUMNS 12
#define  MAX_LEN  256
#define  MAX_O_TERMS  50
#define  BAD_MASK  7  
#define  PI  3.141592653589
#define  MAX_V  3.5
#define  MIN_V  -3.5
#define  MAX_TIME_JUMP  1.9
#define  CHOP_TOL0  4.0
#define  CHOP_SCALE  1.0
#define  MAX 1000
#define  SIGMA 0.5
#define  MIN_TAU 0.015
#define  MIN_STAU 0.005
#define  MIN_WEIGHT 1.0
#define A225 31.3
#define B225 0.014
#define    Amid(x,lo,hi)  (((x)>=(lo))&&((x)<=(hi)))
double chauvenet(long N);
int sigma(double sumsq,double sum, int n, double *mean, double *var);
int read_float_column(fitsfile *fptr, char *name, long nrows, long *ncols,float *array);
int read_double_column(fitsfile *fptr, char *name, long nrows, long *ncols,double *array);
int read_logical_column(fitsfile *fptr, char *name, long nrows, long *ncols,char *array);
void printabort(char *message);
void printerror(int status);
void check_saturation(float *rawdata, int baddet[], long ndet, long nsamp,char *badsamp,char quietflag);
void check_timing(double *ut, double *det_time, long nsamp, char *badsamp,float sampdur);
void calc_tran(float skytran[], long nsamp, char *badsamp, float el[],float tauzen);
void calc_chop(float *chop, long nsamp, float chop_tol, float *chop_hi,float *chop_lo, float *chop_eff);
void chop_demod(float *chop, long *nsamp, float chop_hi, float chop_lo,float chop_tol, char *badsamp, double **rawdata, long ndet,float **el, float **ra, float **dec, float **azerr,float **elerr, float **azo, float **elo, float **rao,float **deco, float **pa, char **on_beam,char **off_beam);
double Gain_Tau(double tau, float t_ele);
float tausmooth(int yyyymmdd0, double ut0, FILE* fp);
void linefit(float *xlist,float *ylist,int nsamp,float *intercept,float *slope);
double median(double *val,int count);

int  IM_X=201, IM_Y=201;
float  IM_X0=101.0, IM_Y0=101.0;
float  ROTATOR_0 = 16.0;
float  MAGNIFICATION = 2.0;    /* map pixels/detector */
char  map_too_small = 'n';
float  A_CEN_X, A_CEN_Y;
float  SCALE = SCALE_CASS;     /* arcsec/detector */
char Usage[]="Usage: sharpinteg file.fits -r rgmfile {-c -f # -t # -a # # -q  -o -d str -y # file -nh -nc -sil -fa -hard -ne -w -shn}\n   -c : Use Chauvenet's theorem for deglitching\n   -f # : Manually set gain between h and v arrays\n    -t  # : Manually set chop tolerance\n   -a # # :  X and Y offsets to align V array with H\n    -q : Quicklook mode outputing h and v array images for each hwp position\n   -o : If this flag is present sharpinteg will output calibration/debug values\n    -d str : Program expects to read in string(date) to be printed to ouput text file generaged by -o flag\n    -y str file: Progam will read in string str (yyyymmdd) to be used for calculating smoothed tau using tau data from file\n    -nh: Do not check hwp angles\n  -nc: Do not check nod/chop agreement\n   -sil: Quiet mode (suppresses some messages to screen\n   -fa: Prints out full array\n   -hard: Do not process chop/nod\n  -ne: No EDAS so do not split into multiple angles based on EDAS Status field\n   -w:  Do Set flux of widow pixels to 2*their flux)\n -shn #:  Minimum number of samples required for a hwp integration\n -em :  Inflate Signal Errors less than median Signal Error to the median\n -et :  Inflate Signal Errors less than the Mean Signal Error - 3*Std Dev Signal Error to the Mean Signal Error\n"
;

struct Header {
  char source[256];
  char platform[256];
  char date[256];
  float raep;
  float decep;
  float equinox;
  float rao0;
  float deco0;
  float azo0;
  float zao0;
  float fazo;
  float fzao;
  float rotator;
  float azo_chop;
  float rotator0;
};






int main(int Argc, char *Argv[])
{
  float year,raep,decep,*skytran,tauzen,sampdur;
  float *mean_pa_i, *mean_el_i, *rawdata_i, *el_i, *az_i,*ra_i, *dec_i, *azerr_i, *elerr_i, *azo_i, *elo_i, *rao_i, *deco_i, *pa_i, *nodstate_i, *chop_i, *source_state_i;
  double *poldata[MAX_HWP],hwpang[MAX_HWP],hwpangcheck;
  double *left[MAX_HWP],*leftsq[MAX_HWP],*right[MAX_HWP],*rightsq[MAX_HWP],*rightlow,*righthigh,*leftlow,*lefthigh,*raw[MAX_HWP];
  double A_CEN_X, A_CEN_Y,A_CEN_X_H,A_CEN_X_V;
  float *offset[MAX_O_TERMS];
  float mask_value = MASK_VALUE0;
  float  chop_throw = -1.0, chop_tol = CHOP_TOL0,chop_hi, chop_lo, chop_eff,val;
  float f[ARRAY_ROWS*SHARP_COLUMNS],fmed,ftemp[ARRAY_ROWS*SHARP_COLUMNS],fset,*xlist,*ylist,*sigx,*sigy,slope[MAX_HWP][SHARP_COLUMNS*ARRAY_ROWS],intercept,sigs,sigi,chi2,z,tempslope,tau225gh;
  float *mean_pa[MAX_HWP],*mean_el[MAX_HWP],mean_az,*rawdata[MAX_HWP],*el[MAX_HWP],*az[MAX_HWP],*ra[MAX_HWP],*dec[MAX_HWP],*azerr[MAX_HWP],*elerr[MAX_HWP],*azo[MAX_HWP],*elo[MAX_HWP],*rao[MAX_HWP],*deco[MAX_HWP],*pa[MAX_HWP],*nodstate[MAX_HWP],*chop[MAX_HWP],*source_state[MAX_HWP],*hwpangle,*hwpstatus;
  double *ut[MAX_HWP],*det_time[MAX_HWP];
  double *polsig[MAX_HWP],*sig,*polerr[MAX_HWP],*sigerr,*demod[MAX_HWP],*demoderr[MAX_HWP],*sigsumerr,*u,*q,*uerr,*qerr,*rgm,*errsigerr;
  double *ut_i, *det_time_i,leftmean,leftvar,rightmean,rightvar,rthres=3.0,lthres=3.0,ut_save[4],errmean,errvar,errmedian,errsum,errsumerr;
  float *h[MAX_HWP],*herr[MAX_HWP],*hwt[MAX_HWP],*v[MAX_HWP],*verr[MAX_HWP],*vwt[MAX_HWP],powoutH,powoutV,*weights;
  long row,nrow,nsamp,ndet,width,ncol0,det,samp,*leftcount[MAX_HWP],*rightcount[MAX_HWP],npdet,nppdet,naxes[3],nelements,fpixel[3],hwpsamp[MAX_HWP],hwpstart[MAX_HWP],samptot,filenum=0,*weight[MAX_HWP],errcount;
  char *off_beam_i, *on_beam_i,*badsamp,commandline[512],line[512];
  char *on_beam[MAX_HWP],*off_beam[MAX_HWP];
  char headline[512],platform[512],nasmyth_flag,source[512],*extname,headline2[256],date[20],*taufile,*junk,morejunk[256];
  char *filename,*rgmfile,deglitchflag=' ',fflag=' ',qflag='n',outputflag='n',tauflag=' ',hfile[MAX_HWP][20],vfile[MAX_HWP][20],syscom[256],hwpcheckflag='y',nccheckflag='y',quietflag='n',fullarrayflag='n',fileout[256],hardflag='n',edasflag='y',widowflag='n',errthreesigflag='n',errmedianflag='n';
  int  status = 0,extver,hdunum,hdutype,firstdatahdu,naxis,bitpix;   /* must be initialized to zero! */
  int colnum,typecode,term,baddet[ARRAY_ROWS*ARRAY_COLUMNS+1],nfiles;
  int i,j,k,ndetcol,ndetrow,arg,x,y,*bad,*badsharp,xoff=0,yoff=0,yyyymmdd0,nhwp,nodcheckl1,nodcheckr,nodcheckl2,samhwpnum=100;
  u_int err;
  FILE *rgmf,*fout,*tablef,*tauf;
  fitsfile  *infptr,*outfptr;
  struct Header head;

  float testarray[144];
  if(Argc < 4) printabort(Usage);
  fprintf(stderr,"This is sharp integ version 2.5 - experimental 2009 version by GN\n");
  strcpy(commandline,Argv[0]);
  for(i=0;i<Argc;i++)
    {
      sprintf(line," %s",Argv[i]);
      strcat(commandline,line);
    }
  filename=Argv[1];
  junk=strstr(filename,"sharc2");
  sscanf(junk,"sharc2-0%d.fits",&filenum);
  fprintf(stderr,"%s %d\n",filename,filenum);
  arg=2;
  if(strncmp(Argv[arg],"-r",2)==0)
    {
      arg++;
      rgmf=fopen(rgmfile=Argv[arg],"r");
      fprintf(stderr,"%s\n",Argv[arg]);
      arg++;
    }
  else printabort(Usage);
  while(arg<Argc)
    {
      if(strncmp(Argv[arg],"-c",2)==0)
	{
	  arg++;
	  deglitchflag='c';
	}
      else if(strncmp(Argv[arg],"-f",2)==0)
	{
	  if(strncmp(Argv[arg],"-fa",3)==0)
	    {
	      arg++;
	      fullarrayflag='y';
	    }
	  else
	    {
	      arg++;
	      fflag='y';
	      sscanf(Argv[arg],"%f\n",&fset);
	      arg++;
	    }
	}
      else if(strncmp(Argv[arg],"-t",2)==0)
	{
	  arg++;
	  sscanf(Argv[arg],"%f\n",&chop_tol);
	  arg++;
	}
      else if(strncmp(Argv[arg],"-m",2)==0)
	{
	  arg++;
	  sscanf(Argv[arg],"%f\n",&mask_value);
	  arg++;
	}
      else if(strncmp(Argv[arg],"-a",2)==0)
	{
	  arg++;
	  sscanf(Argv[arg],"%d\n",&xoff);
	  arg++;
	  sscanf(Argv[arg],"%d\n",&yoff);
	  arg++;
	}
      else if (strncmp(Argv[arg],"-q",2)==0)
	{
	  arg++;
	  qflag='y';
	}
      else if (strncmp(Argv[arg],"-d",2)==0)
	{
	  arg++;
	  sscanf(Argv[arg],"%s",&date);
	  arg++;
	}
      else if(strncmp(Argv[arg],"-o",4)==0)
	{
	  arg++;
	  outputflag='y';
	}
      else if(strncmp(Argv[arg],"-y",2)==0)
	{
	  tauflag='y';
	  arg++;
	  sscanf(Argv[arg],"%d",&yyyymmdd0);
	  arg++;
	  taufile=Argv[arg];
	  arg++;
	}
      else if(strncmp(Argv[arg],"-nh",3)==0)
	{
	  hwpcheckflag='n';
	  arg++;
	}
      else if(strncmp(Argv[arg],"-nc",3)==0)
	{
	  nccheckflag='n';
	  arg++;
	}
      else if(strncmp(Argv[arg],"-sil",4)==0)
	{
	  quietflag='y';
	  arg++;
	}
      else if(strncmp(Argv[arg],"-hard",5)==0)
	{
	  hardflag='y';
	  arg++;
	}
      else if(strncmp(Argv[arg],"-ne",3)==0)
	{
	  edasflag='n';
	  arg++;
	}
      else if(strncmp(Argv[arg],"-w",2)==0)
	{
	  widowflag='y';
	  arg++;
	}
      else if(strncmp(Argv[arg],"-shn",4)==0)
	{
	  arg++;
	  sscanf(Argv[arg],"%d",&samhwpnum);
	  arg++;
	}
      else if(strncmp(Argv[arg],"-em",3)==0)
	{
	  errmedianflag='y';
	  arg++;
	}
      else if(strncmp(Argv[arg],"-et",3)==0)
	{
	  errthreesigflag='y';
	  arg++;
	}
      else 
	{
	  printabort(Usage);
	}
    }
  if(outputflag == 'y')
    {
      tablef=fopen("planet_data.txt","a");
      /*first item in output file*/
      fprintf(tablef,"%s\t",filename);
      /*second item in output file*/
      fprintf(tablef,"%s\t",date);
    }
  if (tauflag == 'y') tauf=fopen(taufile,"r");

  fprintf(stderr, "Reading raw data...%s\n",filename);      fits_open_file(&infptr, filename, READONLY, &status);

  /* Read HDU 0 */
  if (fits_read_key(infptr, TSTRING, "OBJECT", head.source, headline, &status))
    {
      printerror(status);
    }
  nasmyth_flag = 'n';
  if (fits_read_key(infptr, TSTRING, "PLATFORM", head.platform, headline, &status))
    {
      printerror(status);
    }
  else 
    {
      fprintf(stderr,"%s",head.platform);
      if (strncmp(platform, "NA", 2) == 0) 
	{
	  nasmyth_flag = 'y';
	  fprintf(stderr, "  Nasmyth platform detected\n");
	  SCALE = SCALE_NAS;     /* only one SCALE allowed per reduction */
	}
    }
  if (fits_read_key(infptr, TSTRING, "DATE-OBS", head.date, headline, &status))
    {
      printerror(status);
    }
  if (sscanf(head.date, "%d", &year) != 1)
    {
      fprintf(stderr,"Could not read date from FITS file");
    }  
  if (fits_read_key(infptr, TFLOAT, "RAEP", &head.raep, headline, &status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "DECEP", &head.decep, headline, &status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "EQUINOX", &head.equinox, headline, &status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "RAO", &head.rao0, headline,&status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "DECO", &head.deco0, headline,&status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "AZO", &head.azo0, headline,&status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "ZAO", &head.zao0, headline,&status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "FAZO", &head.fazo, headline,&status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "FZAO", &head.fzao, headline,&status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "ROTATOR", &head.rotator, headline,&status))
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TFLOAT, "AZO_CHOP", &head.azo_chop,headline, &status)) 
    {
      printerror(status);
      head.azo_chop = 0.0;
    }
  if (nasmyth_flag == 'y') head.rotator = 0.0;
  if (fits_read_key(infptr, TFLOAT, "ROTZERO", &head.rotator0, headline,&status)) 
    {   /* Seems to handle string keyword OK */
      status = 0;     /* November 2002 data doesn't have ROTZERO in header */
    }
  else 
    {
      ROTATOR_0 = head.rotator0;
    }
  if ((head.rotator < -360.0) || (head.rotator > 360.0)) 
    {
      head.rotator = ROTATOR_0;
      fprintf(stderr, "  WARNING:  Bad rotator value, setting to ");
      system("date >> sharpsolve_errlog.txt");
      system("echo \"WARNING:  Bad rotator value\" >> sharpsolve_errlog.txt");
    }
  head.rotator = ROTATOR_0-head.rotator;
  fprintf(stderr, "  Cal. inst. rot. = %6.1f (file %d)\n", head.rotator,nfiles+1);
  hdunum = 5;
  if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) != 0)
    {
      printerror(status);
    }
  if (fits_read_key(infptr, TSTRING, "EXTNAME", headline, headline2,&status))
    {
      printerror(status);
    }
  firstdatahdu = 5;
  if (strncmp(headline, "SHARC2 Data", 11) != 0)
    {
      firstdatahdu = 6;     /* hdunum = 5 is presumably CSO DSOS */
    }
  /* count total number of frames */
  hdunum = firstdatahdu;
  row = 0;
  while (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) == 0) 
    {
      if (hdutype != BINARY_TBL)
	{
	  printabort("Data HDU of wrong type");
	}
      if (fits_get_num_rows(infptr, &nrow, &status))
	{
	  printerror(status);
	}
      row += nrow;
      hdunum++;
    }
  nsamp = row;
  fprintf(stderr,"total number of rows in file is %d\n",nsamp);
  status = 0;
  /* allocate memory */
  hdunum = firstdatahdu;
  if (fits_movabs_hdu(infptr, hdunum, &hdutype, &status))
    {
      printerror(status);
    }
  if (hdutype != BINARY_TBL)
    {
      printabort("Data HDU of wrong type");
    }
  if (fits_get_colnum(infptr, CASEINSEN, "Pixel Readout", &colnum, &status))
    {
      printerror(status);
    }
  if (fits_get_coltype(infptr, colnum, &typecode, &ndet, &width, &status))
    {
      printerror(status);
    }
  if (ndet != (ARRAY_ROWS*ARRAY_COLUMNS))
    {
      printabort("Unexpected number of detectors");
    }

  rgm = (double *) malloc(ndet*sizeof(double));
  bad=(int *) malloc(ndet*sizeof(int));

      
  for(j=0;j<ndet;j++)
    {
      fscanf(rgmf,"%d %d %lf %d %*lf\n",&x,&y,&(rgm[j]),&(bad[j]));
      if((x-1)*32+y != j+1) 
	{
	  fprintf(stderr,"flaw in reading rgm file %d %d %d %d\n",x,y,(x-1)*32+y,j+1);
	}
    }
  fclose(rgmf);

  if(ndet==384 && fullarrayflag=='n') npdet=288;
  else if (fullarrayflag=='y') npdet=384;
  else fprintf(stderr,"unknown number of polarimeter pixels\n");


  mean_az=0.0; 
  mean_pa_i = (float *) malloc(sizeof(float));
  mean_el_i = (float *) malloc(sizeof(float));
  rawdata_i = (float *) malloc(nsamp*ndet*sizeof(float));
  az_i = (float *) malloc(nsamp*sizeof(float));
  el_i = (float *) malloc(nsamp*sizeof(float));
  ra_i = (float *) malloc(nsamp*sizeof(float));
  dec_i = (float *) malloc(nsamp*sizeof(float));
  azerr_i = (float *) malloc(nsamp*sizeof(float));
  elerr_i = (float *) malloc(nsamp*sizeof(float));
  azo_i = (float *) malloc(nsamp*sizeof(float));
  elo_i = (float *) malloc(nsamp*sizeof(float));
  rao_i = (float *) malloc(nsamp*sizeof(float));
  deco_i = (float *) malloc(nsamp*sizeof(float));
  pa_i = (float *) malloc(nsamp*sizeof(float));
  on_beam_i = (char *) malloc(nsamp*sizeof(char));
  off_beam_i = (char *) malloc(nsamp*sizeof(char));
  nodstate_i = (float *) malloc(nsamp*sizeof(float));
  chop_i = (float *) malloc(nsamp*sizeof(float));
  source_state_i = (float *) malloc(nsamp*sizeof(float));
  ut_i = (double *) malloc(nsamp*sizeof(double));
  det_time_i = (double *) malloc(nsamp*sizeof(double));
  hwpangle = (float *) malloc(nsamp*sizeof(float));
  hwpstatus = (float *) malloc(nsamp*sizeof(float));
  if (fits_read_key(infptr, TFLOAT, "CRPIX3", &A_CEN_Y, headline,&status))
    {
      /*This is only for JSharc data, I believe*/
      fprintf(stderr,"here %lf\n", A_CEN_Y);
      status = 0;
      A_CEN_X = (ARRAY_COLUMNS - 1.0)/2.0;
      A_CEN_Y = (ARRAY_ROWS - 1.0)/2.0;
    }
  else 
    {
      A_CEN_Y -= 1.0;
      if (fits_read_key(infptr, TFLOAT, "CRPIX2", &A_CEN_X, headline,
			&status))
	printabort("Trouble reading CRPIX2");
      A_CEN_X -= 1.0;
    }
  A_CEN_X_H = (SHARP_COLUMNS - 1.0)/2.0;
  A_CEN_X_V = (ARRAY_COLUMNS - 1.0) - (SHARP_COLUMNS - 1.0)/2.0;
  fprintf(stderr, "  Reference X pixels:  %.1f, %.1f\n", A_CEN_X_V+1,
	  A_CEN_X_H+1);

  /*read data*/ 
  hdunum = firstdatahdu;
  row = 0;
  while (fits_movabs_hdu(infptr, hdunum, &hdutype, &status) == 0) 
    {
      if (hdutype != BINARY_TBL) printabort("Data HDU of wrong type");
      if (fits_get_num_rows(infptr, &nrow, &status)) printerror(status);
      if (read_float_column(infptr, "Pixel Readout", nrow, &ncol0,&(rawdata_i[row*ndet]))) 
	printerror(status);
      if (read_float_column(infptr, "AZ", nrow, &ncol0, &(az_i[row])))
	printerror(status);
      if (read_float_column(infptr, "EL", nrow, &ncol0, &(el_i[row]))) 
	printerror(status);
      if (read_float_column(infptr, "RA", nrow, &ncol0, &(ra_i[row]))) 
	printerror(status);
      if (read_float_column(infptr, "DEC", nrow, &ncol0,&(dec_i[row]))) 
	printerror(status);
      if (read_float_column(infptr, "AZ_ERROR", nrow, &ncol0,&(azerr_i[row]))) 
	printerror(status);
      if (read_float_column(infptr, "EL_ERROR", nrow, &ncol0,&(elerr_i[row]))) 
	printerror(status);
      if (read_float_column(infptr, "AZO", nrow, &ncol0,&(azo_i[row])))	    
	printerror(status);
      if (read_float_column(infptr, "ELO", nrow, &ncol0,&(elo_i[row])))
	printerror(status);
      if (read_float_column(infptr, "RAO", nrow, &ncol0,&(rao_i[row])))
        printerror(status);
      if (read_float_column(infptr, "DECO", nrow, &ncol0,&(deco_i[row])))
	printerror(status);
      if (read_float_column(infptr, "Parallactic Angle", nrow, &ncol0,&(pa_i[row])))
	printerror(status);
      if (read_logical_column(infptr, "On Beam", nrow, &ncol0,&(on_beam_i[row])))
	printerror(status);
      if (read_logical_column(infptr, "Off Beam", nrow, &ncol0,&(off_beam_i[row])))
	printerror(status);
      if (read_float_column(infptr, "CHOP_OFFSET", nrow, &ncol0,&(chop_i[row])))
	printerror(status);
      if (read_double_column(infptr, "UT", nrow, &ncol0, &(ut_i[row])))
	printerror(status);
      if (read_double_column(infptr, "Detector Time", nrow, &ncol0,&(det_time_i[row]))) 
	printerror(status);
      if(read_float_column(infptr,"HWP_ANGLE",nrow,&ncol0,&(hwpangle[row])))
	printerror(status);
      if(read_float_column(infptr,"EDAS_STATUS",nrow,&ncol0,&(hwpstatus[row])))
	printerror(status);
      row += nrow;
      hdunum++;
    }
  status = 0;
  if (fits_read_key(infptr, TFLOAT, "CDELT1", &sampdur, headline, &status))
    printerror(status);
  sampdur /= 1000.0;
  fprintf(stderr, "finished source = %s\n",head.source);
  if (fits_close_file(infptr, &status))
    printerror(status);
  
  if (tauflag == 'y') 
    {
      tau225gh=tausmooth(yyyymmdd0,ut_i[nsamp/2],tauf);
      fprintf(stderr,"tau smoothed value is %lf\n",tau225gh);
    }

  /*split input array into arrays for each hwp position*/
  i=0;
  nhwp=0;
  for(samp=0;samp<nsamp;samp++)
    {
      if (hwpstatus[samp] <= 2.2 && hwpstatus[samp] >= 1.8 ) 
	{
	  i++;
	  hwpangcheck+=hwpangle[samp];
	}
      else
	{
	  if (i != 0 && i >= samhwpnum)
	    {
	      hwpsamp[nhwp]=i;
	      hwpstart[nhwp]=samp-i;
	      hwpang[nhwp]=hwpangle[samp-1];
	      if(abs(hwpang[nhwp]-hwpangcheck/hwpsamp[nhwp]) >= 5) fprintf(stderr,"Final angle %lf and average angle %lf do not agree\n",hwpang[nhwp],hwpangcheck/hwpsamp[nhwp]);
	      nhwp++;
	      fprintf(stderr,"%d %d\n",i,nhwp);
	      i=0;
	      hwpangcheck=0;
	    }
	  else
	    {
	      i=0;
	      hwpangcheck=0;
	    }
	}
    }
  if(i != 0 && i >= samhwpnum)
    {
      hwpsamp[nhwp]=i;
      hwpstart[nhwp]=nsamp-i;
      hwpang[nhwp]=hwpangle[nsamp-1];
      if(abs(hwpang[nhwp]-hwpangcheck/hwpsamp[nhwp]) >= 5) fprintf(stderr,"Final angle %lf and average angle %lf do not agree\n",hwpang[nhwp],hwpangcheck/hwpsamp[nhwp]);
      nhwp++;
    }
  if(edasflag=='n')
    {
      nhwp=1;
      hwpsamp[nhwp-1]=nsamp;
      hwpstart[nhwp-1]=0;
    }
  fprintf(stderr,"%d %d\n",i,nhwp);
  fprintf(stderr,"Number of hwp's in file is %d\n",nhwp);
  for (j=0;j<nhwp;j++) fprintf(stderr,"Number of SHARC-II samples in hwp are %d and begining SHARC-II sample is %d\n",hwpsamp[j],hwpstart[j]);
  for (samp=0;samp<nsamp;samp++) 
    {	
      *(mean_el_i) += el_i[samp]/nsamp;
      mean_az+=az_i[samp]/nsamp;
    }
  for(j=0;j<nhwp;j++)
    {
      fprintf(stderr,"got here\n");
      mean_pa[j] = (float *) malloc(sizeof(float));
      mean_el[j] = (float *) malloc(nhwp*sizeof(float));
      rawdata[j] = (float *) malloc(nsamp*ndet*sizeof(float));
      az[j] = (float *) malloc(nsamp*sizeof(float));
      el[j] = (float *) malloc(nsamp*sizeof(float));
      ra[j] = (float *) malloc(nsamp*sizeof(float));
      dec[j] = (float *) malloc(nsamp*sizeof(float));
      azerr[j] = (float *) malloc(nsamp*sizeof(float));
      elerr[j] = (float *) malloc(nsamp*sizeof(float));
      azo[j] = (float *) malloc(nsamp*sizeof(float));
      elo[j] = (float *) malloc(nsamp*sizeof(float));
      rao[j] = (float *) malloc(nsamp*sizeof(float));
      deco[j] = (float *) malloc(nsamp*sizeof(float));
      pa[j] = (float *) malloc(nsamp*sizeof(float));
      on_beam[j] = (char *) malloc(nsamp*sizeof(char));
      off_beam[j] = (char *) malloc(nsamp*sizeof(char));
      nodstate[j] = (float *) malloc(nsamp*sizeof(float));
      chop[j] = (float *) malloc(nsamp*sizeof(float));
      source_state[j] = (float *) malloc(nsamp*sizeof(float));
      ut[j] = (double *) malloc(nsamp*sizeof(double));
      det_time[j] = (double *) malloc(nsamp*sizeof(double));



      for(i=0,samptot=hwpstart[j];i<hwpsamp[j];i++,samptot++)
	{
	  for (det=0;det<ndet;det++) rawdata[j][i*ndet+det]=rawdata_i[samptot*ndet+det];
	  az[j][i]=az_i[samptot];
	  el[j][i]=el_i[samptot];
	  ra[j][i]=ra_i[samptot];
	  dec[j][i]=dec_i[samptot];
	  azerr[j][i]=azerr_i[samptot];
	  elerr[j][i]=elerr_i[samptot];
	  azo[j][i]=azo_i[samptot];
	  elo[j][i]=elo_i[samptot];
	  rao[j][i]=rao_i[samptot];
	  deco[j][i]=deco_i[samptot];
	  pa[j][i]=pa_i[samptot];
	  on_beam[j][i]=on_beam_i[samptot];
	  off_beam[j][i]=off_beam_i[samptot];
	  nodstate[j][i]=nodstate_i[samptot];
	  chop[j][i]=chop_i[samptot];
	  source_state[j][i]=source_state_i[samptot];
	  ut[j][i]=ut_i[samptot];
	  det_time[j][i]=det_time_i[samptot];
	}
      ut_save[j]=ut[j][0];

      badsamp = (char *) malloc(nsamp * sizeof(char));
      skytran = (float *) malloc(nsamp * sizeof(float));
      calc_tran(skytran, hwpsamp[j],badsamp, el[j],tauzen);
      for (samp=0;samp<hwpsamp[j];samp++) badsamp[samp]=0;
      /*check a few things (stolen from sharpsolve may not really need)*/
      check_saturation(rawdata[j], baddet, ndet, hwpsamp[j],badsamp,quietflag);
      check_timing(ut[j], det_time[j], hwpsamp[j], badsamp, sampdur);
      free (ut[j]);
 fprintf(stderr,"through loop time %d\n",j);      
    }

  free(rawdata_i);
  free(mean_pa_i);
  free(az_i);
  free(el_i);
  free(ra_i);
  free(dec_i);
  free(azerr_i);
  free(elerr_i);
  free(azo_i);
  free(elo_i);
  free(rao_i);
  free(deco_i);
  free(pa_i);
  free(on_beam_i);
  free(off_beam_i);
  free(nodstate_i);
  free(chop_i);
  free(source_state_i);
  free(ut_i);
  free(det_time_i);
  free(hwpangle);
  free(hwpstatus);
/*demodulate the chop*/
  for(j=0;j<nhwp;j++)
    {
      calc_chop(chop[j], hwpsamp[j], chop_tol, &chop_hi, &chop_lo,&chop_eff);
      fprintf(stderr,"Measured chop high, low (eff.):  %.1f, %.1f  (%.f%%)\n",chop_hi, chop_lo, chop_eff*100.0);
      fprintf(stderr, "Measured throw = %.1f arcsec\n",(chop_hi-chop_lo)*CHOP_SCALE);
      if(abs((chop_hi-chop_lo)*CHOP_SCALE-2*abs(head.azo_chop)) > 2)
	{
	  fprintf(stderr,"Measured chopper throw and nod length %lf not compatible\n",2*head.azo_chop);
	  if (nccheckflag=='y') return(0);
	}
      raw[j]=(double *)malloc(hwpsamp[j]*ndet*sizeof(double));
     
      for(samp=0;samp<hwpsamp[j]*ndet;samp++){
	
	if(bad[samp%ndet] !=0)
	  {
	    raw[j][samp]=0;
	    /*if((samp%ndet)%ARRAY_ROWS < 12) rawdata[nfiles][samp+20]=0;
	      else if((samp%ndet)%ARRAY_ROWS >=20) rawdata[nfiles][samp-20]=0;*/
	  }
	else raw[j][samp]=(double)-1*rawdata[j][samp]/rgm[samp%ndet];

      }
      free(rawdata[j]);
      if(hardflag=='n') chop_demod(chop[j], &(hwpsamp[j]), chop_hi, chop_lo, chop_tol,badsamp, &(raw[j]), ndet, &(el[j]),&(ra[j]), &(dec[j]), &(azerr[j]), &(elerr[j]),&(azo[j]), &(elo[j]), &(rao[j]), &(deco[j]),&(pa[j]), &(on_beam[j]), &(off_beam[j]));

      if(outputflag == 'y')
	{
	  /*Take mean of elevation angle data same thing is normally done in cal_rot_pixel_xy)*/
	  for (samp=0;samp<hwpsamp[j];samp++)
	    *(mean_el[j]) += el[j][samp];
	  *(mean_el[j]) /= hwpsamp[j];
	  fprintf(stderr,"Tele_elv:  %f\n",(*mean_el));
	  /*Print to output txt file*/
	  fprintf(tablef,"%f\t",(*mean_el));
	}
       
      xlist= (float *) malloc((hwpsamp[j]+1)*sizeof(float));
      ylist= (float *) malloc((hwpsamp[j]+1)*sizeof(float));
      sigx = (float *) malloc((hwpsamp[j]+1)*sizeof(float));
      sigy = (float *) malloc((hwpsamp[j]+1)*sizeof(float));
    
      fprintf(stderr,"Number of Frames are %d for hwp %d\n",hwpsamp[j],j);

      for(ndetrow=0;ndetrow<12;ndetrow++)
	{
	  for(ndetcol=0;ndetcol<12;ndetcol++)
	    {
	      if(ndetcol+xoff<12 && ndetcol+xoff>=0 && ndetrow+yoff<12 && ndetrow+yoff>=0) {	 
		for (samp=0;samp<hwpsamp[j];samp++)
		  {
		    det=ndetrow*32+ndetcol;
		   
		    
		    xlist[samp]=raw[j][samp*ndet+ndetrow*32+ndetcol];
		    ylist[samp]=raw[j][samp*ndet+(ndetrow+yoff)*32+ndetcol+xoff+20];
		    /*values don't matter*/
		 
		    sigx[samp+1]=1.0e5;
		    sigy[samp+1]=1.0e5;
		  
		  }
	      
		slope[j][ndetrow*12+ndetcol]=0;
		if(bad[ndetrow*32+ndetcol]==0 && bad[(ndetrow+yoff)*32+ndetcol+20+xoff]==0) linefit(xlist,ylist,hwpsamp[j],&intercept,&slope[j][ndetrow*12+ndetcol]);
	      }
	    }
	} 
      free(xlist);
      free(ylist);
      free(sigx);
      free(sigy);

      
      poldata[j] = (double *) malloc(nsamp*npdet*sizeof(double));
    }

  badsharp=(int *) malloc(npdet*sizeof(int));
  for(det=0;det<npdet/2;det++)
    {
      for(i=0;i<3;i++)
	{
	  for(j=i+1;j<4;j++)
	    {
	      if(slope[j][det]<slope[i][det])
		{
		  tempslope=slope[i][det];
		  slope[i][det]=slope[j][det];
		  slope[j][det]=tempslope;
		}
	    }
	}
    }     
  if(strncmp(&fflag,"y",1)!=0) 
    {
      for(det=0;det<npdet/2;det++)
	{
	  f[det]=(slope[1][det]+slope[2][det])/2;
	  fprintf(stdout,"%d %f\n",det,f[det]);
	}
    }
  else 
    {
      for(det=0;det<npdet/2;det++)
	{
	  f[det]=fset;
	}
    }

  for(ndetrow=0;ndetrow<12;ndetrow++)
    {
      for(ndetcol=0;ndetcol<12;ndetcol++)
	{
	  badsharp[12*ndetrow+ndetcol]=bad[ndetrow*32+ndetcol];
	  badsharp[12*ndetrow+ndetcol+144]=bad[ndetrow*32+ndetcol+20];
	}
    }


  for(i=0;i<nhwp;i++)
    {
      for(samp=0;samp<hwpsamp[i];samp++)
	{
	  if(fullarrayflag=='y')
	    {
	      for(det=0;det<ndet;det++)
		{
		  poldata[i][samp*ndet+det]=raw[i][samp*ndet+det];
		}
	    }
	  else
	    {
	      for(ndetrow=0;ndetrow<12;ndetrow++)
		{
		  for(ndetcol=0;ndetcol<12;ndetcol++)
		    {
		      if(qflag=='y')
			{
			  poldata[i][samp*npdet+12*ndetrow+ndetcol]=raw[i][samp*ndet+ndetrow*32+ndetcol];
			  poldata[i][samp*npdet+12*ndetrow+ndetcol+npdet/2]=raw[i][samp*ndet+ndetrow*32+ndetcol+20];
			}
		      else if(ndetcol+xoff<12 && ndetcol+xoff>=0 && ndetrow+yoff<12 && ndetrow+yoff>=0)
			{
			  poldata[i][samp*npdet+12*ndetrow+ndetcol]=raw[i][samp*ndet+ndetrow*32+ndetcol]-raw[i][samp*ndet+(ndetrow+yoff)*32+ndetcol+xoff+20]/f[12*ndetrow+ndetcol];
			  poldata[i][samp*npdet+12*ndetrow+ndetcol+npdet/2]=raw[i][samp*ndet+ndetrow*32+ndetcol]+raw[i][samp*ndet+(ndetrow+yoff)*32+ndetcol+xoff+20]/f[12*ndetrow+ndetcol];
			}
		      else
			{
			  badsharp[12*ndetrow+ndetcol]=10;
			  poldata[i][samp*288+12*ndetrow+ndetcol]=0;
			  poldata[i][samp*288+12*ndetrow+ndetcol+144]=0;
			}
		    }
		}
	    }
	}
      free(raw[i]);
    }


  nppdet=npdet/2;
  for(i=0;i<nhwp;i++)
    {

      left[i] = (double *)malloc(npdet*sizeof(double));
      right[i] = (double *)malloc(npdet*sizeof(double)); 
      leftsq[i] = (double *)malloc(npdet*sizeof(double));
      rightsq[i] = (double *)malloc(npdet*sizeof(double));
      polsig[i]=(double *)malloc(nppdet*sizeof(double));
      polerr[i]=(double *)malloc(nppdet*sizeof(double));
      demod[i]=(double *)malloc(npdet*sizeof(double));
      demoderr[i]=(double *)malloc(npdet*sizeof(double));
      q=(double *)malloc(npdet/2*sizeof(double));
      u=(double *)malloc(npdet/2*sizeof(double));
      qerr=(double *)malloc(npdet/2*sizeof(double));
      uerr=(double *)malloc(npdet/2*sizeof(double));
      leftcount[i]=(long *)malloc(npdet*sizeof(long));
      rightcount[i]=(long *)malloc(npdet*sizeof(long));
      leftlow=(double *)malloc(npdet*sizeof(double));
      lefthigh=(double *)malloc(npdet*sizeof(double));
      rightlow=(double *)malloc(npdet*sizeof(double));
      righthigh=(double *)malloc(npdet*sizeof(double));
      weight[i]=(long *)malloc(npdet*sizeof(long));
      weights=(float *)malloc(npdet*sizeof(float));
      if(fullarrayflag=='y')
	{
	  sig=(double *)malloc(npdet*sizeof(double));
	  sigerr=(double *)malloc(npdet*sizeof(double));
	  sigsumerr=(double *)malloc(npdet*sizeof(double));
	}
      else
	{
	  sig=(double *)malloc(nppdet*sizeof(double));
	  sigerr=(double *)malloc(nppdet*sizeof(double));
	  sigsumerr=(double *)malloc(nppdet*sizeof(double));
	}

      for (det=0;det<npdet;det++)
	{
	  left[i][det]=0;
	  leftsq[i][det]=0;
	  leftcount[i][det]=0;
	  leftlow[det]=0;
	  lefthigh[det]=0;
	  right[i][det]=0;
	  rightsq[i][det]=0;
	  righthigh[det]=0;
	  rightlow[det]=0;
	  rightcount[i][det]=0;
	  demod[i][det]=0;
	  demoderr[i][det]=0;
	  nodcheckl1=0;
	  nodcheckl2=0;
	  nodcheckr=0;
	  weight[i][det]=0;
	}

      /*add left beams and right beams*/
      for(samp=0;samp<hwpsamp[i];samp++)
	{
	  if(on_beam[i][samp] || hardflag=='y') 
	    {
	      for(det=0;det<npdet;det++)
		{
		  left[i][det]+=poldata[i][samp*npdet+det];
		  leftsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		  leftcount[i][det]+=1;
		}
	    }
	  else if(off_beam[i][samp])
	    {
	      for(det=0;det<npdet;det++)
		{
		  nodcheckl1=leftcount[i][det];
		  right[i][det]+=poldata[i][samp*npdet+det];
		  rightsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		  rightcount[i][det]+=1;
		}
	    }
	}
      nodcheckr=rightcount[i][12];
      nodcheckl2=leftcount[i][12]-nodcheckl1;
      fprintf(stderr,"Nod check: First left has %d Frames Right has %d Frames and Second left has %d\n",nodcheckl1,nodcheckr,nodcheckl2);
      if(deglitchflag=='c')
	{
	  rthres=chauvenet(rightcount[i][0]);
	  lthres=chauvenet(leftcount[i][0]);
	}
      fprintf(stderr,"thresholds are %lf %lf\n",rthres,lthres);
      /*process left and right beams to gather statistics for deglitching*/
      for (det=0;det<npdet;det++)
	{
	  err+=sigma(leftsq[i][det],left[i][det],leftcount[i][det],&leftmean,&leftvar);
	  leftlow[det]=leftmean-lthres*sqrt(leftvar);
	  lefthigh[det]=leftmean+lthres*sqrt(leftvar);
	  
	  err+=sigma(rightsq[i][det],right[i][det],rightcount[i][det],&rightmean,&rightvar);
	  rightlow[det]=rightmean-rthres*sqrt(rightvar);
	  righthigh[det]=rightmean+rthres*sqrt(rightvar);
	  
	}
      for(det=0;det<npdet;det++)
	{      
	  left[i][det]=0;
	  leftsq[i][det]=0;
	  leftcount[i][det]=0;
	  right[i][det]=0;
	  rightsq[i][det]=0;
	  rightcount[i][det]=0;
	}
      /*combine deglitched nodpairs*/
      for(samp=0;samp<hwpsamp[i];samp++)
	{
	  if(on_beam[i][samp] || hardflag=='y') 
	    {
	      for(det=0;det<npdet;det++)
		{
		  if(Amid(poldata[i][samp*npdet+det],leftlow[det],lefthigh[det]))
		    {
		      /*if(det==43) fprintf(stdout,"left %lf %lf\n",poldata[i][samp*npdet+det],poldata[i][samp*npdet+det+144]);*/
		      left[i][det]+=poldata[i][samp*npdet+det];
		      leftsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		      leftcount[i][det]++;
		    }
		  else if(quietflag=='n')
		    {
		      if(qflag=='y')
			{
			  if(det <144) fprintf(stderr,"Deglitch of Frame %d H pixel %d of file %i\n",samp+1,det+1,i+1);
			  else fprintf(stderr,"Deglitch of Frame %d V pixel %d of file %i\n",samp+1,det-143,i+1);
			}
		      else
			{
			  if(det <144) fprintf(stderr,"Deglitch of Frame %d H-V pixel %d of file %i\n",samp+1,det+1,i+1);
			  else fprintf(stderr,"Deglitch of Frame %d H+V pixel %d of file %i\n",samp+1,det-143,i+1);
			}
		    }
		}
	    }
	  else if(off_beam[i][samp])
	    {
	      for(det=0;det<npdet;det++)
		{
		  if(Amid(poldata[i][samp*npdet+det],rightlow[det],righthigh[det]))
		    {
		      /*if(det==43) fprintf(stdout,"right %lf %lf\n",poldata[i][samp*npdet+det],poldata[i][samp*npdet+det+144]);*/
		      right[i][det]+=poldata[i][samp*npdet+det];
		      rightsq[i][det]+=poldata[i][samp*npdet+det]*poldata[i][samp*npdet+det];
		      rightcount[i][det]++;
		    }
		  else if(quietflag=='n')
		    {
		      if(qflag=='y')
			{
			  if(det <144) fprintf(stderr,"Deglitch of Frame %d H pixel %d of file %i\n",samp+1,det+1,i+1);
			  else fprintf(stderr,"Deglitch of Frame %d V pixel %d of file %i\n",samp+1,det-143,i+1);
			}
		      else
			{
			  if(det <144) fprintf(stderr,"Deglitch of Frame %d H-V pixel %d of file %i\n",samp+1,det+1,i+1);
			  else fprintf(stderr,"Deglitch of Frame %d H+V pixel %d of file %i\n",samp+1,det-143,i+1);
			}
		    }
		}
	    }
	}
      free(poldata[i]);
    }




  for(i=0;i<nhwp;i++)
    {
      for(det=0;det<npdet/2;det++)
	{
	  sig[det]=0;
	  sigsumerr[det]=0;
	  weights[det]=0;
	  if(fullarrayflag=='y')
	    {
	      sig[det+npdet/2]=0;
	      sigsumerr[det+npdet/2]=0;
	    }
	  polsig[i][det]=0;
	  polerr[i][det]=0;
	}
    }
/*quicklook addition*/
  if(qflag=='y')
    {
      for(i=0;i<nhwp;i++)
	{
	  h[i]=(float *)malloc(npdet/2*sizeof(float));
	  v[i]=(float *)malloc(npdet/2*sizeof(float));
	  herr[i]=(float *)malloc(npdet/2*sizeof(float));
	  verr[i]=(float *)malloc(npdet/2*sizeof(float));
	  hwt[i]=(float *)malloc(npdet/2*sizeof(float));
	  vwt[i]=(float *)malloc(npdet/2*sizeof(float));
	}
    }
  free(leftlow);
  free(lefthigh);
  free(righthigh);
  free(rightlow);
    
  for(i=0;i<nhwp;i++)
    {
      for(det=0;det<npdet;det++)
	{
	  err+=sigma(leftsq[i][det],left[i][det],leftcount[i][det],&leftmean,&leftvar);
	  err+=sigma(rightsq[i][det],right[i][det],rightcount[i][det],&rightmean,&rightvar);
	  demod[i][det]=leftmean-rightmean;
	  if (hardflag == 'n') demoderr[i][det]=sqrt(leftvar/leftcount[i][det]+rightvar/rightcount[i][det]);
	  else
	    {
	      if(rightcount[i][det] >=0) demoderr[i][det]=sqrt(rightvar/rightcount[i][det]);
	      else if (leftcount[i][det] >= 0) demoderr[i][det]=sqrt(leftvar/leftcount[i][det]);
	      else fprintf(stdout,"File shows samples in both left and right positions\n");
	    }
	  weight[i][det]=leftcount[i][det]+rightcount[i][det];
	}
      free(left[i]);
      free(right[i]);
      free(leftsq[i]);
      free(rightsq[i]);
      free(leftcount[i]);
      free(rightcount[i]);
      
      if(fullarrayflag=='y')
	{
	  for(det=0;det<npdet;det++)
	    {
	      sig[det]+=demod[i][det]/4.0;
	      sigsumerr[det]+=demod[i][det]*demod[i][det];
	    }
	}

      else if(qflag=='y')
	{
	  for(det=0;det<npdet/2;det++)
	    {
	      h[i][det]=0;
	      herr[i][det]=0;
	      v[i][det]=0;
	      verr[i][det]=0;
	      h[i][det]=demod[i][det];
	      herr[i][det]=demoderr[i][det];
	      hwt[i][det]=1.0;
	      v[i][det]=demod[i][det+npdet/2];
	      verr[i][det]=demoderr[i][det+npdet/2];
	      vwt[i][det]=1.0;	
	    }
	}
      else 
	{
	  for(det=0;det<npdet/2;det++)
	    {
	      polsig[i][det]=demod[i][det];
	      polerr[i][det]=demoderr[i][det];
	      sig[det]+=demod[i][det+npdet/2]/4;
	      sigsumerr[det]+=demod[i][det+npdet/2]*demod[i][det+npdet/2];
	      weights[det]+=weight[i][det];
	    }
	}
    }

  if(qflag=='y')
    {
      for(det=0;det<npdet/2;det++)
	{
	  if(badsharp[det] != 0)	    {
	      for(i=0;i<nhwp;i++)
		{
		  h[i][det]=NAN;
		  herr[i][det]=NAN;
		  hwt[i][det]=0.0;
		}
	    }
	  if(badsharp[det+npdet/2] != 0)
	    {
	      for(i=0;i<nhwp;i++)
	      {
		v[i][det]=NAN;
		verr[i][det]=NAN;
		vwt[i][det]=0.0;
	      }
	    }
	}
      for(i=0;i<nhwp;i++)
	{

	  sprintf(hfile[i],"h_%d_%d.fits",filenum,i);
	  sprintf(vfile[i],"v_%d_%d.fits",filenum,i);
	  sprintf(syscom,"rm -f %s",hfile[i]);
	  system(syscom);
	  sprintf(syscom,"rm -f %s",vfile[i]);
	  system(syscom);

	  if(outputflag == 'y')
	    {
	      /*Writing output file: arrays "h" and "v" hold data*/
	      fprintf(tablef,"%f\t",tau225gh);
	      fprintf(tablef,"%lf\t",ut_save[j]);
	      for (j=0;j<npdet/2;j++)
		{
		  if(h[i][j]>=0)
		    powoutH+=h[i][j];
		  if(v[i][j]>=0)
		    powoutV+=v[i][j];
		}
	      fprintf(tablef,"%f\t",powoutH);
	      fprintf(tablef,"%f\t",powoutV);
	      /*Now apply corrections*/
	      powoutH=0;
	      powoutV=0;
	      for(j=0;j<npdet/2;j++)
		{
		  h[i][j]=Gain_Tau(tau225gh,(*mean_el[i]))*h[i][j];
		  v[i][j]=Gain_Tau(tau225gh,(*mean_el[i]))*v[i][j];
		  if(h[i][j]>=0)
		    powoutH+=h[i][j];
		  if(v[i][j]>=0)
		    powoutV+=v[i][j];
		}
	      fprintf(tablef,"%f\t",powoutH);
	      fprintf(tablef,"%f\t",powoutV);
	      fprintf(tablef,"\n");
	    }
	      
	  /* WRITE OUTPUT FILES */
	  bitpix = FLOAT_IMG;     /* changed 2006 Jul 3 */
	  naxis = 2;
	  naxes[0] = 12;
	  naxes[1] = 12;
	  nelements = 144;
	  fpixel[0] = 1;
	  fpixel[1] = 1;

      /* CREATE H FILE */
	 
	  if (fits_create_file(&outfptr, hfile[i], &status))
	    printerror(status);

	  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	    printerror(status);
	  if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, h[i], &status))
	    printerror(status);  
	  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "H SIGNAL","", &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "SOURCE", &(head.source),"",&status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "DATE", &(head.date),"",&status))
	    printerror(status);
	  if (fits_write_key(outfptr, TFLOAT, "EQUINOX", &(head.equinox), "", &status))
	    printerror(status);
	  val=*(mean_el[i]);     
	  if (fits_write_key(outfptr,TFLOAT, "TEL_ELEV", &val, "", &status))
	    printerror(status);
	  val=pa[i][hwpsamp[i]/2]; 
	  if (fits_write_key(outfptr,TFLOAT, "PARA_ANG", &val, "", &status))		printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "CTYPE1", "ALON-SIN", "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CRVAL1", &(val), "", &status))
	    printerror(status);
	  val = SCALE/3600.0;
	  if (fits_write_key(outfptr, TFLOAT, "CDELT1", &val, "", &status))
	    printerror(status);
	  val=6.5;
	  if (fits_write_key(outfptr, TFLOAT, "CRPIX1", &val, "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CROTA1", &val, "", &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "CTYPE2", "ALAT-SIN", "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CRVAL2", &(val), "", &status))
	    printerror(status);
	  val=SCALE/3600.0;
	  if (fits_write_key(outfptr, TFLOAT, "CDELT2", &val, "", &status))
	    printerror(status);
	  val=6.5;
	  if (fits_write_key(outfptr, TFLOAT, "CRPIX2", &val, "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CROTA2", &val, "", &status))
	    printerror(status);
	  val=head.raep;
	  if (fits_write_key(outfptr, TFLOAT, "RA", &val, "", &status))
	    printerror(status);
	  val=head.decep;
	  if (fits_write_key(outfptr, TFLOAT, "DEC", &val, "", &status))
	    printerror(status);
	  val=head.rao0;
	  if (fits_write_key(outfptr, TFLOAT, "RAO", &val, "", &status))
	    printerror(status);
	  val=head.deco0;
	  if (fits_write_key(outfptr, TFLOAT, "DECO", &val, "", &status))
	    printerror(status);
	  val=head.azo0;
	  if (fits_write_key(outfptr, TFLOAT, "AZO", &val, "", &status))
	    printerror(status);
	  val=head.zao0;
	  if (fits_write_key(outfptr, TFLOAT, "ZAO", &val, "", &status))
	    printerror(status);
	  val=head.rotator;
	  if(fits_write_key(outfptr,TFLOAT,"DEWARANG",&val,"",&status))
	    printerror(status);
	  sprintf(line,"Created using sharpinteg version 2.5");
	  if(fits_write_comment(outfptr,line,&status))
	    printerror(status);
	  if(fits_write_comment(outfptr,commandline,&status))
	    printerror(status);
	  if(fits_write_date(outfptr,&status))
	    printerror(status);
	  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "H WEIGHT","", &status))
	    printerror(status);
	  if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, hwt[i], &status))
	    printerror(status);    

	  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "H ERROR","", &status))
	    printerror(status);
	  if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, herr[i], &status))
	    printerror(status);    
	  if (fits_open_file(&infptr, filename, READONLY, &status))
	    printerror(status);
	  if (fits_copy_hdu(infptr, outfptr, 0, &status))
	    printerror(status);
	  if (fits_close_file(infptr, &status))
	    printerror(status);

	  if(fits_close_file(outfptr,&status))
	    printerror(status);
	  
      /* CREATE V FILE */
	  	        
	  if (fits_create_file(&outfptr, vfile[i], &status))
	    printerror(status);

	  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	    printerror(status);
	  if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, v[i], &status))
	    printerror(status);  
	  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "V SIGNAL","", &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "SOURCE", &(head.source),"",&status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "DATE", &(head.date),"",&status))
	    printerror(status);
	  if (fits_write_key(outfptr, TFLOAT, "EQUINOX", &(head.equinox), "", &status))
	    printerror(status);
	  val=*(mean_el[i]);     
	  if (fits_write_key(outfptr,TFLOAT, "TEL_ELEV", &val, "", &status))
	    printerror(status);
	  val=pa[i][hwpsamp[i]/2]; 
	  if (fits_write_key(outfptr,TFLOAT, "PARA_ANG", &val, "", &status))		printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "CTYPE1", "ALON-SIN", "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CRVAL1", &(val), "", &status))
	    printerror(status);
	  val = SCALE/3600.0;
	  if (fits_write_key(outfptr, TFLOAT, "CDELT1", &val, "", &status))
	    printerror(status);
	  val=6.5;
	  if (fits_write_key(outfptr, TFLOAT, "CRPIX1", &val, "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CROTA1", &val, "", &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "CTYPE2", "ALAT-SIN", "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CRVAL2", &(val), "", &status))
	    printerror(status);
	  val=SCALE/3600.0;
	  if (fits_write_key(outfptr, TFLOAT, "CDELT2", &val, "", &status))
	    printerror(status);
	  val=6.5;
	  if (fits_write_key(outfptr, TFLOAT, "CRPIX2", &val, "", &status))
	    printerror(status);
	  val=0.0;
	  if (fits_write_key(outfptr, TFLOAT, "CROTA2", &val, "", &status))
	    printerror(status);
	  val=head.raep;
	  if (fits_write_key(outfptr, TFLOAT, "RA", &val, "", &status))
	    printerror(status);
	  val=head.decep;
	  if (fits_write_key(outfptr, TFLOAT, "DEC", &val, "", &status))
	    printerror(status);
	  val=head.rao0;
	  if (fits_write_key(outfptr, TFLOAT, "RAO", &val, "", &status))
	    printerror(status);
	  val=head.deco0;
	  if (fits_write_key(outfptr, TFLOAT, "DECO", &val, "", &status))
	    printerror(status);
	  val=head.azo0;
	  if (fits_write_key(outfptr, TFLOAT, "AZO", &val, "", &status))
	    printerror(status);
	  val=head.zao0;
	  if (fits_write_key(outfptr, TFLOAT, "ZAO", &val, "", &status))
	    printerror(status);
	  val=head.rotator;
	  if(fits_write_key(outfptr,TFLOAT,"DEWARANG",&val,"",&status))
	    printerror(status);
	  sprintf(line,"Created using sharpinteg version 2.1");
	  if(fits_write_comment(outfptr,line,&status))
	    printerror(status);
	  if(fits_write_comment(outfptr,commandline,&status))
	    printerror(status);
	  if(fits_write_date(outfptr,&status))
	    printerror(status);
	  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "V WEIGHT","", &status))
	    printerror(status);
	  if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, vwt[i], &status))
	    printerror(status);    

	  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	    printerror(status);
	  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "V ERROR","", &status))
	    printerror(status);
	  if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, verr[i], &status))
	    printerror(status);    
	  if (fits_open_file(&infptr, filename, READONLY, &status))
	    printerror(status);
	  if (fits_copy_hdu(infptr, outfptr, 0, &status))
	    printerror(status);
	  if (fits_close_file(infptr, &status))
	    printerror(status);

	  if(fits_close_file(outfptr,&status))
	    printerror(status);
	    
	}
      return(0);
    }

	/* 111  */
	
  if(fullarrayflag=='y')
    {
      for(det=0;det<npdet;det++)
	{
	  sigerr[det]=sqrt(fabs(sigsumerr[det]-sig[det]*sig[det]*4)/3);
	  if(bad[det]!=0)
	    {
	      sig[det]=NAN;
	      sigerr[det]=NAN;
	    }
	}
      sprintf(fileout,"%d.fits",filenum);
      sprintf(morejunk,"rm -f %s",fileout);
      system(morejunk);
      if (fits_create_file(&outfptr, fileout, &status))
	printerror(status);
      bitpix = DOUBLE_IMG;
      naxis = 2;
      naxes[0] = 32;
      naxes[1] = 12;
      nelements = 384;
      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	printerror(status);
      fpixel[0] = 1;
      fpixel[1] = 1;
      if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, sig, &status))
	printerror(status);
      if (fits_write_key(outfptr, TSTRING, "EXTNAME", "SIGNAL","", &status))
	printerror(status);
      if (fits_write_key(outfptr, TSTRING, "SOURCE", &(head.source),"",&status))
	printerror(status);
      if (fits_write_key(outfptr, TSTRING, "DATE", &(head.date),"",&status))
	printerror(status);
      if (fits_write_key(outfptr, TFLOAT, "EQUINOX", &(head.equinox), "", &status))
	printerror(status);
      val=mean_az;
      if (fits_write_key(outfptr,TFLOAT,"TEL_AZ",&val,"",&status))
	printerror(status);
      val=*(mean_el_i);
      if (fits_write_key(outfptr,TFLOAT,"TEL_ELEV",&val,"",&status))
	printerror(status);
      val=pa[2][0];
      if (fits_write_key(outfptr,TFLOAT,"PARA_ANG",&val,"",&status))
	printerror(status);
      if (fits_write_key(outfptr, TSTRING, "CTYPE1", "ARRAY-X", "", &status))
	printerror(status);
      val=6.5;
      if (fits_write_key(outfptr, TFLOAT, "CRVAL1", &(val), "", &status))
	printerror(status);
      val = SCALE/-3600.0;
      /*val=1;*/
      if (fits_write_key(outfptr, TFLOAT, "CDELT1", &val, "", &status))
	printerror(status);
      val=6.5;
      if (fits_write_key(outfptr, TFLOAT, "CRPIX1", &val, "", &status))
	printerror(status);
      val=0.0;
      if (fits_write_key(outfptr, TFLOAT, "CROTA1", &val, "", &status))
	printerror(status);
      if (fits_write_key(outfptr, TSTRING, "CTYPE2", "ARRAY-Y", "", &status))
	printerror(status);
      val=6.5;
      if (fits_write_key(outfptr, TFLOAT, "CRVAL2", &(val), "", &status))
	printerror(status);
      val = SCALE/3600.0;
	/*val=1;*/
      if (fits_write_key(outfptr, TFLOAT, "CDELT2", &val, "", &status))
	printerror(status);
      val=6.5;
      if (fits_write_key(outfptr, TFLOAT, "CRPIX2", &val, "", &status))
	printerror(status);
      val=0.0;
      if (fits_write_key(outfptr, TFLOAT, "CROTA2", &val, "", &status))
	printerror(status);
      val=head.raep;
      if (fits_write_key(outfptr, TFLOAT, "RA", &val, "", &status))
	printerror(status);
      val=head.decep;
      if (fits_write_key(outfptr, TFLOAT, "DEC", &val, "", &status))
	printerror(status);
      val=head.rao0;
      if (fits_write_key(outfptr, TFLOAT, "RAO", &val, "", &status))
	printerror(status);
      val=head.deco0;
      if (fits_write_key(outfptr, TFLOAT, "DECO", &val, "", &status))
	printerror(status);
      val=head.azo0;
      if (fits_write_key(outfptr, TFLOAT, "AZO", &val, "", &status))
	printerror(status);
      val=head.zao0;
      if (fits_write_key(outfptr, TFLOAT, "ZAO", &val, "", &status))
	printerror(status);
      val=head.rotator;
      if (fits_write_key(outfptr,TFLOAT, "DEWARANG", &val, "", &status))
	printerror(status);
      val=0.0;
      if(fits_write_key(outfptr,TFLOAT,"CAZO",&val,"",&status))
	printerror(status);
      if(fits_write_key(outfptr,TFLOAT,"CZAO",&val,"",&status))
	printerror(status);
      sprintf(line,"Created using sharpinteg version 2.1");
      if(fits_write_comment(outfptr,line,&status))
	printerror(status);
      if(fits_write_comment(outfptr,commandline,&status))
	printerror(status);
      if(fits_write_date(outfptr,&status))
	printerror(status);
      if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
	printerror(status);
      if (fits_write_key(outfptr, TSTRING, "EXTNAME", "SIGNAL ERROR","", &status))
	printerror(status);
      if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, sigerr, &status))
        printerror(status);
      if (fits_close_file(outfptr, &status))
	printerror(status);

      return(0);
    }
	

  /* 222  */

  if(nhwp != 4  && edasflag!='n')
    {
      fprintf(stderr,"Number of hwp positions %d is not equal to 4\n",nhwp);
      return(0);
    }

  for(det=0;det<npdet/2;det++)  
    {
      if(edasflag!='n')
        {  
          sigerr[det]=sqrt(fabs(sigsumerr[det]-sig[det]*sig[det]*4)/3);
	  q[det]=0.5*(polsig[0][det]-polsig[2][det]);
	  if(hwpang[2]-hwpang[0]-45 <= -10 || hwpang[2]-hwpang[0]-45 >= 10) fprintf(stderr,"Q angles not within 10 degrees of 45\n");
	  u[det]=0.5*(polsig[1][det]-polsig[3][det]);
	  if(hwpang[3]-hwpang[1]-45 <= -10 || hwpang[3]-hwpang[1]-45 >= 10) fprintf(stderr,"U angles not within 10 degrees of 45\n");
	  qerr[det]=0.5*sqrt((polerr[0][det]*polerr[0][det]+polerr[2][det]*polerr[2][det]));
	  uerr[det]=0.5*sqrt((polerr[1][det]*polerr[1][det]+polerr[3][det]*polerr[3][det]));
	}
      else
	{
	  sig[det]*=4;
	  sigerr[det]=demoderr[0][det];
	  q[det]=0;
	  u[det]=0;
	  qerr[det]=0;
	  uerr[det]=0;
	}
    }
 

  /* Here is where it checks for widows, makes badsharp 144 not 288 elements, and checks Q U errors  */

	
  errcount=npdet/2;
  errsigerr=(double *)malloc(errcount*sizeof(double));
  for(det=0;det<npdet/2;det++)
    {
      if(badsharp[det] !=0 || badsharp[det+npdet/2] !=0)
	{
	  if(widowflag != 'y' || badsharp[det]*badsharp[det+npdet/2]!=0)
	    {
	      sig[det]=NAN;
	      sigerr[det]=NAN;
	      weights[det]=0;
	      errcount--;
	    }
	  else
	    {
	      sig[det]=2*sig[det];
	      sigerr[det]=2*sigerr[det];
	    }
	  u[det]=NAN;
	  uerr[det]=NAN;
	  q[det]=NAN;
	  qerr[det]=NAN;
	  badsharp[det]+=badsharp[det+npdet/2];
	  weights[det]/=2.0;
	}
	if ((uerr[det]>=mask_value/1000000.0) || (qerr[det]>=mask_value/1000000.0)) {
		if (badsharp[det]==0) {
			badsharp[det]=1;
		}
		u[det]=NAN;
		uerr[det]=NAN;
		q[det]=NAN;
		qerr[det]=NAN;
	}				
    }


  fprintf(stderr,"Number of good pixels in error mean are %d\n",errcount);


  /* this seems to be median filtering code  */
	
  errsum=0;
  errsumerr=0;
  errmean=0;
  errvar=0;
  errmedian=0;

  k=0;
  for(det=0;det<npdet/2;det++)
      {
	if(weights[det] != 0)
	{
	  errsum+=sigerr[det];
	  errsumerr+=sigerr[det]*sigerr[det];
          errsigerr[k]=sigerr[det];
	  k++;
	}
      }
  err=sigma(errsumerr,errsum,errcount,&errmean,&errvar);
  errmedian=median(errsigerr,errcount);
  for(det=0;det<npdet/2;det++)
      {
	if(weights[det] != 0)
	{
	  if(errthreesigflag=='y') 
	    {
	      if(sigerr[det] < errmean-3*sqrt(errvar)) sigerr[det]=errmean;
		
	    }
	  else if(errmedianflag=='y')
	    {
	      if(sigerr[det] < errmedian) sigerr[det]=errmedian;
	    }
	}
      }
   free(errsigerr);
  sprintf(fileout,"%d_int.fits",filenum);
  fprintf(stderr,"%s\n",fileout);
  sprintf(morejunk,"rm -f %s",fileout);
  system(morejunk);
  if (fits_create_file(&outfptr, fileout, &status))
    printerror(status);
  bitpix = DOUBLE_IMG;
  naxis = 2;
  naxes[0] = 12;
  naxes[1] = 12;
  nelements = 144;
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  fpixel[0] = 1;
  fpixel[1] = 1;
	
  /* The signal array is about to be written  */
	
	
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, sig, &status))
    printerror(status); 
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "SIGNAL","", &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "SOURCE", &(head.source),"",&status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "DATE", &(head.date),"",&status))
    printerror(status);
  if (fits_write_key(outfptr, TFLOAT, "EQUINOX", &(head.equinox), "", &status))
    printerror(status);
  val=mean_az;
  if (fits_write_key(outfptr,TFLOAT,"TEL_AZ",&val,"",&status))
    printerror(status);
  val=*(mean_el_i);
  if (fits_write_key(outfptr,TFLOAT,"TEL_ELEV",&val,"",&status))
    printerror(status);
  if (edasflag == 'n') val=pa[0][0];
  else val=pa[2][0];
  if (fits_write_key(outfptr,TFLOAT,"PARA_ANG",&val,"",&status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "CTYPE1", "ARRAY-X", "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRVAL1", &(val), "", &status))
    printerror(status);
  val = SCALE/-3600.0;
  /*  val=1;*/
  if (fits_write_key(outfptr, TFLOAT, "CDELT1", &val, "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRPIX1", &val, "", &status))
    printerror(status);
  val=0.0;
  if (fits_write_key(outfptr, TFLOAT, "CROTA1", &val, "", &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "CTYPE2", "ARRAY-Y", "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRVAL2", &(val), "", &status))
    printerror(status);
  val = SCALE/3600.0;
  /*val=1;*/
  if (fits_write_key(outfptr, TFLOAT, "CDELT2", &val, "", &status))
    printerror(status);
  val=6.5;
  if (fits_write_key(outfptr, TFLOAT, "CRPIX2", &val, "", &status))
    printerror(status);
  val=0.0;
  if (fits_write_key(outfptr, TFLOAT, "CROTA2", &val, "", &status))
    printerror(status);
  val=head.raep;
  if (fits_write_key(outfptr, TFLOAT, "RA", &val, "", &status))
    printerror(status);
  val=head.decep;
  if (fits_write_key(outfptr, TFLOAT, "DEC", &val, "", &status))
    printerror(status);
  val=head.rao0;
  if (fits_write_key(outfptr, TFLOAT, "RAO", &val, "", &status))
    printerror(status);
  val=head.deco0;
  if (fits_write_key(outfptr, TFLOAT, "DECO", &val, "", &status))
    printerror(status);
  val=head.azo0;
  if (fits_write_key(outfptr, TFLOAT, "AZO", &val, "", &status))
    printerror(status);
  val=head.zao0;
  if (fits_write_key(outfptr, TFLOAT, "ZAO", &val, "", &status))
    printerror(status);
  val=head.rotator;
  if (fits_write_key(outfptr,TFLOAT, "DEWARANG", &val, "", &status))
    printerror(status);
  val=0.0;
  if(fits_write_key(outfptr,TFLOAT,"CAZO",&val,"",&status))
    printerror(status);
  if(fits_write_key(outfptr,TFLOAT,"CZAO",&val,"",&status))
    printerror(status);
  sprintf(line,"Created using sharpinteg version 2.1");
  if(fits_write_comment(outfptr,line,&status))
    printerror(status);
  if(fits_write_comment(outfptr,commandline,&status))
    printerror(status);
  if(fits_write_date(outfptr,&status))
     printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "SIGNAL WEIGHT","", &status))
    printerror(status);
	
  /* weights, Q, U, etc... about to be written */
	
  if (fits_write_pix(outfptr, TFLOAT, fpixel, nelements, weights, &status))
    printerror(status); 
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "SIGNAL ERROR","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, sigerr, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "Q","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, q, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "Q ERROR","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, qerr, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "U","", &status))
    printerror(status);
  for (i=0;i<nppdet;i++) u[i]=-1*u[i];
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, u, &status))
        printerror(status);
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "U ERROR","", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TDOUBLE, fpixel, nelements, uerr, &status))
        printerror(status);
  for (i=0;i<144;i++) testarray[i]=sqrt(u[i]*u[i]+q[i]*q[i])/sig[i];
  if (fits_create_img(outfptr, bitpix, naxis, naxes, &status))
    printerror(status);
  if (fits_write_key(outfptr, TSTRING, "EXTNAME", "PIXEL MASK","", &status))
    printerror(status);
  det=0;
  if (fits_write_key(outfptr, TINT, "VALUE FOR GOOD PIXEL", &det, "", &status))
    printerror(status);
  if (fits_write_pix(outfptr, TINT, fpixel, nelements, badsharp, &status))
  printerror(status);
  if (fits_open_file(&infptr, filename, READONLY, &status))
    printerror(status);
  if (fits_copy_hdu(infptr, outfptr, 0, &status))
    printerror(status);
  if (fits_close_file(infptr, &status))
    printerror(status);


  if (fits_close_file(outfptr, &status))
    printerror(status);
  return 0;
}
/***********************/

int sigma(double sumsq,double sum, int n, double *mean, double *var)
{
  if (n<2)
    {
      *mean=sum;
      *var=0.0;
      return(1);
    }
  else
    {
      *mean=sum/n;
      *var=(sumsq-*mean * *mean * n)/(n-1);
      return(0);
    }
}

/***********************/

double median(double *val,int count)
{
  int i,j;
  double dummy;

  for (i=0;i<count-1;i++)
    {
      for (j=i;j<count;j++)
	{
	  if(val[i] < val[j]) 
	    {
	      dummy=val[i];
	      val[i]=val[j];
	      val[j]=dummy;
	    }
	}
    }
  if (count % 2 == 1) return(val[(count-1)/2]);
  else return(0.5*(val[count/2]+val[count/2-1]));
}
/***********************/
double chauvenet(long N)
{
  double low=0.0,new,high,value,reject,root2,precision;
  double TH_PREC=0.0004;

  reject=1.0/(2.0*N);
  root2=sqrt(2.0);
  precision=TH_PREC*root2;
  high=10.0/root2;

  while( (high-low)>precision)
    {
      new=(low+high)/2.0;
      value=erfc(new);
      if(value>reject)
	low=new;
      else
	high=new;
    }
  return((low+high)/root2);
}

/***********************/
int read_float_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                      float *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    float  nulval = FLOATNULLVALUE;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TFLOAT) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(float)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}
/*******************/

int read_double_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                       double *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    double  nulval = DOUBLENULLVALUE;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TDOUBLE) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(double)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}

/*******************/

int read_logical_column(fitsfile *fptr, char *name, long nrows, long *ncols,
                        char *array)

{
    int  colnum;
    int  status = 0;
    int  typecode;
    long  width;
    char  message[MAX_LEN];
    long  firstrow, firstelem, nelements;
    char  nulval = 0;
    int  anynul;

    if (fits_get_colnum(fptr, CASEINSEN, name, &colnum, &status))
      return (status);
    if (fits_get_coltype(fptr, colnum, &typecode, ncols, &width, &status))
      printerror(status);
    if (typecode != TLOGICAL) {
      sprintf(message, "%s data of wrong type", name);
      printabort(message);
    }
    if (width != sizeof(char)) {
      sprintf(message, "%s data of wrong size", name);
      printabort(message);
    }
    firstrow = 1;
    firstelem = 1;
    nelements = nrows*(*ncols);
    if (fits_read_col(fptr, typecode, colnum, firstrow, firstelem,
                      nelements, &nulval, array, &anynul, &status))
      printerror(status);
    return (0);
}

/*******************/

void printabort(char *message)

{
    fprintf(stderr, "ERROR:  %s\n", message);
    system("date >> sharpsolve_errlog.txt");
    system("echo \"ABORTING:  unrecoverable error\" >> sharpsolve_errlog.txt");
    exit (1);
}

/*******************/

void printerror(int status)

{
    if (status)
    {
       fits_report_error(stderr, status);
       system("date >> sharpsolve_errlog.txt");
       system("echo \"ABORTING:  FITS error\" >> sharpsolve_errlog.txt");
       exit (status);
    }
    return;
}

/*******************/
void calc_tran(float skytran[], long nsamp, char *badsamp, float el[],
               float tauzen)

{
    long  samp;

    for (samp = 1; samp <= nsamp; samp++) {
      skytran[samp-1] = exp(-1.0*tauzen/sin(el[samp-1]*PI/180.0));
      if (el[samp-1] < 0.1) {
        fprintf(stderr, "  WARNING:  Bad elevation in frame %d -- flagging\n",
                samp);
        badsamp[samp-1] = 1;
        skytran[samp-1] = 1.0e-6;
        system("date >> sharpsolve_errlog.txt");
        system("echo \"WARNING:  Bad elevation\" >> sharpsolve_errlog.txt");
      }
    }
}

/*******************/

void check_saturation(float *rawdata, int baddet[], long ndet, long nsamp,
                      char *badsamp, char quietflag)

{
    long  det, samp;
    char  ok_flag[ARRAY_ROWS*ARRAY_COLUMNS];
    float  raw;

    for (det = 1; det <= ndet; det++)
      ok_flag[det-1] = 'y';
    for (det = 1; det <= ndet; det++) {
      for (samp = 1; samp <= nsamp; samp++) {
        if (badsamp[samp-1] == 0) {
          raw = rawdata[(samp-1)*ndet + (det-1)];
          if ((raw > MAX_V) || (raw < MIN_V))
            ok_flag[det-1] = 'n';
        }
      }
    }
    for (det = 1; det <= ndet; det++) {
      if ((ok_flag[det-1] != 'y') && ((baddet[det] & BAD_MASK) == 0)) {
        if(quietflag=='n') fprintf(stderr,
                "  WARNING:  Detector (%2d, %2d) near saturation\n",
                (det-1)/ARRAY_COLUMNS+1, (det-1)%ARRAY_COLUMNS+1);
        system("date >> sharpsolve_errlog.txt");
        system(
        "echo \"WARNING:  Detector near saturation\" >> sharpsolve_errlog.txt");
      }
    }
}

/*******************/

void check_timing(double *ut, double *det_time, long nsamp, char *badsamp,
                  float sampdur)

{
    long  samp, minsamp, maxsamp;
    double  dut, mindut, maxdut;

    mindut = 1.0e30;
    maxdut = -1.0e30;
    for (samp = 1; samp < nsamp; samp++) {
      if ((badsamp[samp-1] == 0) && (badsamp[samp] == 0)) {
        dut = ut[samp] - ut[samp-1];
        if (dut < mindut) {
          mindut = dut;
          minsamp = samp+1;
        }
        if (dut > maxdut) {
          maxdut = dut;
          maxsamp = samp+1;
        }
      }
    }
    if (mindut < 0.0) {
      fprintf(stderr, "  WARNING:  Time not monotonic for sample %ld\n",
              minsamp);
      system("date >> sharpsolve_errlog.txt");
      system("echo \"WARNING:  Time not monotonic\" >> sharpsolve_errlog.txt");
    }
    maxdut *= 3600.0;
    if (maxdut > (MAX_TIME_JUMP*sampdur)) {
      fprintf(stderr, "  WARNING:  Time jump (%.3lf sec) for sample %ld\n",
              maxdut, maxsamp);
      system("date >> sharpsolve_errlog.txt");
      system("echo \"WARNING:  Time jump\" >> sharpsolve_errlog.txt");
    }
    mindut = 1.0e30;
    maxdut = -1.0e30;
    if (det_time[0] > -1.0e29) {
      for (samp = 1; samp <= nsamp; samp++) {
        if ((badsamp[samp-1] == 0)) {
          dut = det_time[samp-1] - ut[samp-1]*3600.0;     /* seconds */
          if (dut < mindut) {
            mindut = dut;
            minsamp = samp+1;
          }
          if (dut > maxdut) {
            maxdut = dut;
            maxsamp = samp+1;
          }
        }
      }
      dut = (maxdut + mindut)/2.0;
      dut = dut - 86400.0*((int) (dut/86400.0 + 0.5));
      fprintf(stderr, "  Post April 2003 timing check:  %.3f %.3f %.3f %.3f\n",
              mindut, maxdut, dut, maxdut - mindut);
      if (fabs(dut) > (MAX_TIME_JUMP*sampdur)) {
        fprintf(stderr,
          "  WARNING:  Synchronization offset:  %.3f sec\n", dut);
        system("date >> sharpsolve_errlog.txt");
        system(
         "echo \"WARNING:  Synchronization offset\" >> sharpsolve_errlog.txt");
      }
      if ((maxdut - mindut) > (MAX_TIME_JUMP*sampdur)) {
        fprintf(stderr,
          "  WARNING:  Synchronization glitch:  %.3f sec\n", maxdut-mindut);
        system("date >> sharpsolve_errlog.txt");
        system(
         "echo \"WARNING:  Synchronization glitch\" >> sharpsolve_errlog.txt");
      }
    }
}

/*******************/

void calc_chop(float *chop, long nsamp, float chop_tol, float *chop_hi,
               float *chop_lo, float *chop_eff)

{
    float  himean, lomean, hisig, losig;
    float  himax, himin, lomax, lomin;
    float  mean, sum0, hi0, lo0;
    long  samp;
    int  i;

    /* Should I exclude bad samples? */
    mean = 0.0;
    sum0 = 0.0;
    for (samp = 1; samp <= nsamp; samp++) {
      mean += chop[samp-1];
      sum0 += 1.0;
    }
    mean /= sum0;
    himax = 1.0e30;
    himin = mean;
    lomin = -1.0e30;
    lomax = mean;
    for (i = 1; i <= 10; i++) {
      himean = 0.0;
      lomean = 0.0;
      hi0 = 0.0;
      lo0 = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
       if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax)) {
          himean += chop[samp-1];
          hi0 += 1.0;
        }
        if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax)) {
          lomean += chop[samp-1];
          lo0 += 1.0;
        }
      }
      himean /= hi0;
      lomean /= lo0;
      hisig = 0.0;
      losig = 0.0;
      hi0 = 0.0;
      lo0 = 0.0;
      for (samp = 1; samp <= nsamp; samp++) {
        if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax)) {
          hisig += (chop[samp-1] - himean)*(chop[samp-1] - himean);
          hi0 += 1.0;
        }
        if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax)) {
          losig += (chop[samp-1] - lomean)*(chop[samp-1] - lomean);
          lo0 += 1.0;
        }
      }
      hisig = sqrt(hisig/hi0);
      losig = sqrt(losig/lo0);
      if (hisig < (chop_tol / 2.0))
        hisig = chop_tol / 2.0;
      if (losig < (chop_tol / 2.0))
        losig = chop_tol / 2.0;
      himax = himean + hisig;
      himin = himean - hisig;
      lomax = lomean + losig;
      lomin = lomean - losig;
    }
    *chop_hi = himean;
    *chop_lo = lomean;
    *chop_eff = (hi0 + lo0)/sum0;
    /* assumes that sigmas eventually become smaller than chop_tol / 2.0 */
}

/*******************/

void chop_demod(float *chop, long *nsamp, float chop_hi, float chop_lo,
                float chop_tol, char *badsamp, double **rawdata, long ndet,
                float **el, float **ra, float **dec, float **azerr,
                float **elerr, float **azo, float **elo, float **rao,
                float **deco, float **pa,  char **on_beam, char **off_beam)

{
    long  *choptag;
    float  *hiweight, *loweight;
    float  himax, himin, lomax, lomin;
    long  hicount, locount;
    int  old_chop_state, new_chop_state,beamcheck;
    long  samp, nchop, samp0, samp1;
    double  *c_rawdata;
    float  *c_el, *c_ra, *c_dec, *c_azerr, *c_elerr;
    float  *c_azo, *c_elo, *c_rao, *c_deco, *c_pa; 
    char *c_off,*c_on;
    long  det;

/*     for (samp=1; samp <= *nsamp; samp++) */
/*       fprintf(stderr,"%d, %lf\n", samp, (*nodstate)[samp-1]); */

    /* Doesn'T CORRECTLY USE SCANNING FLAG */
    /* SHOULD HANDLE PA DIFFERENTLY */

    choptag = (long *) malloc(*nsamp*sizeof(long));
    hiweight = (float *) malloc(*nsamp*sizeof(float));
    loweight = (float *) malloc(*nsamp*sizeof(float));

    himax = chop_hi + chop_tol/2.0;
    himin = chop_hi - chop_tol/2.0;
    lomax = chop_lo + chop_tol/2.0;
    lomin = chop_lo - chop_tol/2.0;
    fprintf(stderr,"chop tolerances:himax %f himin %f lomax %f lomin %f\n",himax,himin,lomax,lomin);
    hicount = 0;
    locount = 0;
    old_chop_state = 0;
    nchop = 0;
    samp0 = 1;
    for (samp = 1; samp <= *nsamp; samp++) {
      new_chop_state = 0;
      hiweight[samp-1] = 0.0;
      loweight[samp-1] = 0.0;
      if ((chop[samp-1] >= himin) && (chop[samp-1] <= himax))
        new_chop_state = 1;
      if ((chop[samp-1] >= lomin) && (chop[samp-1] <= lomax))
        new_chop_state = -1;
      if((*on_beam)[samp-1]==(*on_beam)[samp0-1] && abs((*elerr)[samp-1]) < 5 && abs((*azerr)[samp-1]) < 5){
	if (new_chop_state != 0) {
	  if ((new_chop_state*old_chop_state) == -1) {
	    if ((locount > (hicount/2)) && (hicount > (locount/2))) {
	      for (samp1 = samp0; samp1 < samp; samp1++) {
		hiweight[samp1-1] /= (float)hicount;
		loweight[samp1-1] /= (float)locount;
	      }
	      hicount = 0;
	      locount = 0;
	      nchop++;
	      samp0 = samp;
	    }
	    else {
	      for (samp1 = samp0; samp1 < samp; samp1++) {
		if (new_chop_state == 1)
		  hiweight[samp1-1] = 0.0;
		if (new_chop_state == -1)
		  loweight[samp1-1] = 0.0;
	      }
	      if (new_chop_state == 1)
		hicount = 0;
	      if (new_chop_state == -1)
		locount = 0;
	    }
	  }
	  if (badsamp[samp-1] == 0) {
	    if (new_chop_state == 1) {
	      hiweight[samp-1] = 1.0;
	      hicount++;
	    }
	    if (new_chop_state == -1) {
	      loweight[samp-1] = 1.0;
	      locount++;
	    }
	  }
	  old_chop_state = new_chop_state;
	}
      }
      else {
	for(samp1=samp0;samp1<samp;samp1++){
	  hiweight[samp1-1]=0;
	  loweight[samp1-1]=0;
	}
	hicount=0;
	locount=0;
	samp0=samp;
	if(new_chop_state !=0) old_chop_state = new_chop_state;
      }
      choptag[samp-1] = nchop;
    }
    if ((locount > (hicount/2)) && (hicount > (locount/2))) {
      for (samp1 = samp0; samp1 < samp; samp1++) {
        hiweight[samp1-1] /= (float)hicount;
        loweight[samp1-1] /= (float)locount;
      }
      hicount = 0;
      locount = 0;
      nchop++;
      samp0 = samp;
    }

    c_rawdata = (double *) malloc(nchop*ndet*sizeof(double));
    c_el = (float *) malloc(nchop*sizeof(float));
    c_ra = (float *) malloc(nchop*sizeof(float));
    c_dec = (float *) malloc(nchop*sizeof(float));
    c_azerr = (float *) malloc(nchop*sizeof(float));
    c_elerr = (float *) malloc(nchop*sizeof(float));
    c_azo = (float *) malloc(nchop*sizeof(float));
    c_elo = (float *) malloc(nchop*sizeof(float));
    c_rao = (float *) malloc(nchop*sizeof(float));
    c_deco = (float *) malloc(nchop*sizeof(float));
    c_pa = (float *) malloc(nchop*sizeof(float));
    c_on = (char *) malloc(nchop*sizeof(float));
    c_off = (char *) malloc(nchop*sizeof(float));

    for (samp = 1; samp <= nchop; samp++) {
      for (det = 1; det <= ndet; det++) {
        c_rawdata[(samp-1)*ndet + (det-1)] = 0.0;
      }

      c_el[samp-1] = 0.0;
      c_ra[samp-1] = 0.0;
      c_dec[samp-1] = 0.0;
      c_azerr[samp-1] = 0.0;
      c_elerr[samp-1] = 0.0;
      c_azo[samp-1] = 0.0;
      c_elo[samp-1] = 0.0;
      c_rao[samp-1] = 0.0;
      c_deco[samp-1] = 0.0;
      c_pa[samp-1] = 0.0;
      c_on[samp-1] = 0.0;
      c_off[samp-1]=0.0;
    }

    for (samp = 1; samp <= *nsamp; samp++) {
      if (choptag[samp-1] < nchop) {
        for (det = 1; det <= ndet; det++) {
          c_rawdata[choptag[samp-1]*ndet + (det-1)] +=
                                     (*rawdata)[(samp-1)*ndet + (det-1)] *
                                     (hiweight[samp-1]-loweight[samp-1]);
        }
        c_el[choptag[samp-1]] += (*el)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_ra[choptag[samp-1]] += (*ra)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_dec[choptag[samp-1]] += (*dec)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_azerr[choptag[samp-1]] += (*azerr)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_elerr[choptag[samp-1]] += (*elerr)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_azo[choptag[samp-1]] += (*azo)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_elo[choptag[samp-1]] += (*elo)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_rao[choptag[samp-1]] += (*rao)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_deco[choptag[samp-1]] += (*deco)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
        c_pa[choptag[samp-1]] += (*pa)[samp-1] *
                                     (hiweight[samp-1]+loweight[samp-1])/2.0;
 
        c_on[choptag[samp-1]] = (*on_beam)[samp-1];
	c_off[choptag[samp-1]] = (*off_beam)[samp-1];
 	/*fprintf(stdout,"%d, %d, %d, %lf %lf, %lf\n",samp,choptag[samp-1],c_on[choptag[samp-1]],chop[samp-1],hiweight[samp-1], loweight[samp-1]);*/
      }
    }


    free(*rawdata);
    *rawdata = c_rawdata;
    free(*el);
    *el = c_el;
    free(*ra);
    *ra = c_ra;
    free(*dec);
    *dec = c_dec;
    free(*azerr);
    *azerr = c_azerr;
    free(*elerr);
    *elerr = c_elerr;
    free(*azo);
    *azo = c_azo;
    free(*elo);
    *elo = c_elo;
    free(*rao);
    *rao = c_rao;
    free(*deco);
    *deco = c_deco;
    free(*pa);
    *pa = c_pa;
    free(*on_beam);
    *on_beam = c_on;
    free(*off_beam);
    *off_beam=c_off;



    *nsamp = nchop;
    free(choptag);
    free(hiweight);
    free(loweight);

}

/*******************/
double Gain_Tau(double tau, float t_ele)
{
  double M,theta,output;

  theta = PI/2 - (PI*t_ele/180);
  if (t_ele==0)
    M = 100;
  else
    M = 1/(cos(theta));
  
  //output = exp(deltaA*(tau-B225)*M)*exp(A225*(tau-B225)*M);
  output = exp(A225*(tau-B225)*M);
  
  /*if tau = 0 then the Gain_Tau function is disabled*/
  if (tau==0)
    output = 1;
  
  return(output);

}
/*******************/
float tausmooth(int yyyymmdd0, double ut0, FILE* fp)
{
    char  line[MAX];
    int  yyyymmdd, flag;
    float  ut, stau,f,tau;
    double  sum0, sum1, sum2;

    sum0 = 0.0;
    sum1 = 0.0;
    sum2 = 0.0;
    while(fgets(line, MAX, fp) != NULL) {
      if (sscanf(line, "%d %f %f %f", &yyyymmdd, &ut, &tau, &stau) == 4) {
        if (sscanf(line, "%*d %*f %*f %*f %d", &flag) != 1) {
          if (yyyymmdd == yyyymmdd0) {
            if (tau >= MIN_TAU) {
              f = ut - ut0;
              f = exp(-0.5*f*f/(SIGMA*SIGMA));
              if (stau < MIN_STAU)
                stau = MIN_STAU;
              sum0 += 1.0*f;
              sum1 += 1.0*f/(stau*stau);
              sum2 += tau*f/(stau*stau);
            }
          }
        }
      }
    }

    if (sum0 >= MIN_WEIGHT) {
      tau = sum2/sum1;
      fprintf(stderr, "tau225 mean = %6.4f, weight sum = %.3f\n", tau, sum0);
      fprintf(stdout, "%6.4f\n", tau);
    }
    else {
      fprintf(stderr, "Not enough data points; weight sum = %.3f\n", sum0);
      fprintf(stdout, "0.0\n");
    }
	
	return(tau);
}
/*******************/

void linefit(float *xlist,float *ylist,int nsamp,float *intercept,float *slope)
{
     int i;
     double ysum=0,xsum=0,xysum=0,xxsum=0;
     for (i=0;i<nsamp;i++)
       {
	 ysum+=ylist[i];
	 xsum+=xlist[i];
	 xysum+=xlist[i]*ylist[i];
	 xxsum+=xlist[i]*xlist[i];
       }
     *slope=(xysum-xsum*ysum/nsamp)/(xxsum-xsum*xsum/nsamp);
     *intercept=(ysum-*slope*xsum)/nsamp;
}
